""" MultiQC module to parse output from somalier """


import csv
import logging
import random
from collections import defaultdict
from math import isinf, isnan

import spectra

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, heatmap, scatter, table
from multiqc.utils import mqc_colour

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    somalier module class, parses stderr logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Somalier",
            anchor="somalier",
            href="https://github.com/brentp/somalier",
            info="calculates genotype :: pedigree correspondence checks from sketches derived from BAM/CRAM or VCF",
            doi="10.1186/s13073-020-00761-2",
        )

        # Find and load any somalier reports
        self.somalier_data = dict()
        self.somalier_background_pcs = dict()
        self.somalier_ancestry_cats = list()
        self.somalier_length_counts = dict()
        self.somalier_length_exp = dict()
        self.somalier_length_obsexp = dict()

        # parse somalier sample file
        for f in self.find_log_files("somalier/samples"):
            parsed_data = self.parse_somalier_samples(f)
            if parsed_data is not None:
                for s_name_raw in parsed_data:
                    s_name = "*".join([self.clean_s_name(s, f) for s in s_name_raw.split("*")])
                    if s_name in self.somalier_data.keys():
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.somalier_data[s_name] = parsed_data[s_name_raw]

        # parse somalier CSV files
        for f in self.find_log_files("somalier/pairs"):
            parsed_data = self.parse_somalier_pairs_tsv(f)
            if parsed_data is not None:
                for s_name_raw in parsed_data:
                    s_name = "*".join([self.clean_s_name(s, f) for s in s_name_raw.split("*")])
                    if s_name in self.somalier_data.keys():
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.somalier_data[s_name] = parsed_data[s_name_raw]

        # parse somalier ancestry files
        for f in self.find_log_files("somalier/somalier-ancestry", filehandles=True):
            self.parse_somalier_ancestry(f)

        # Filter to strip out ignored sample names
        self.somalier_data = self.ignore_samples(self.somalier_data)

        if len(self.somalier_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.somalier_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.somalier_data, "multiqc_somalier")

        # Somalier Stats Table
        self.somalier_stats_table()

        # Relatedness plot
        self.somalier_relatedness_plot()

        self.somalier_relatedness_heatmap_plot()

        # hetcheck plot
        self.somalier_het_check_plot()

        self.somalier_sex_check_plot()

        # ancestry plots
        self.somalier_ancestry_barplot()

        self.somalier_ancestry_pca_plot()

    @staticmethod
    def parse_somalier_samples(f):
        """Go through log file looking for somalier output"""
        parsed_data = dict()
        headers = None
        sample_i = -100
        for line in f["f"].splitlines():
            s = line.split("\t")
            if headers is None:
                s[0] = s[0].lstrip("#")
                headers = s
                sample_i = headers.index("sample_id")
            else:
                parsed_data[s[sample_i]] = dict()
                for i, v in enumerate(s):
                    if i != sample_i:
                        try:
                            parsed_data[s[sample_i]][headers[i]] = float(v)
                        except ValueError:
                            parsed_data[s[sample_i]][headers[i]] = v
        if len(parsed_data) == 0:
            return None
        return parsed_data

    @staticmethod
    def parse_somalier_pairs_tsv(f):
        """Parse csv output from somalier"""
        parsed_data = dict()
        headers = None
        s_name_idx = None
        for line in f["f"].splitlines():
            s = line.lstrip("#").split("\t")
            if headers is None:
                headers = s
                try:
                    s_name_idx = [headers.index("sample_a"), headers.index("sample_b")]
                except ValueError:
                    log.warning(f"Could not find sample name in somalier output: {f['fn']}")
                    return None
            else:
                s_name = "*".join([s[idx] for idx in s_name_idx])  # not safe to hard code, but works
                parsed_data[s_name] = dict()
                for i, v in enumerate(s):
                    if i not in s_name_idx:  # Skip if (i == 0 or 1); i.e. sample_a, sample_b
                        if isnan(float(v)) or isinf(float(v)):
                            # Inf or NaN indicate the absence of data
                            v = None
                        else:
                            # add the pattern as a suffix to key
                            v = float(v)
                        parsed_data[s_name][headers[i]] = v

        if len(parsed_data) == 0:
            return None
        return parsed_data

    def parse_somalier_ancestry(self, f):
        # dict for parsed data, ancestry prediction probabilities and PCs
        parsed_data = dict()

        # list for background principal components and associated ancestry
        bg_pc1 = []
        bg_pc2 = []
        bg_ancestry = []

        reader: csv.DictReader = csv.DictReader(f["f"], dialect="excel-tab")
        idx = "#sample_id"

        # check file not empty, else parse file
        if reader.fieldnames is not None:
            # get ancestry categories from header fieldnames
            for c in reader.fieldnames:
                # use _prob substring to identify ancestry cats and add to object
                if "_prob" in c:
                    self.somalier_ancestry_cats.append(c.replace("_prob", ""))

            # parse rows of tsv file
            for row in reader:
                # only background has given_ancestry
                # i.e. row is sample, not background
                if len(row["given_ancestry"]) == 0:
                    d = {}
                    for k, v in row.items():
                        # add ancestry prediction to d
                        if "_prob" in k and v is not None:
                            d[k.replace("_prob", "")] = float(v)
                        # add principal component to d
                        elif "PC" in k and v is not None:
                            d[k] = float(v)
                        # else: do nothing

                    # extract predicted ancestry and probability
                    # for general stats table
                    d["ancestry"] = row["predicted_ancestry"]
                    d["p_ancestry"] = d[d["ancestry"]]

                    parsed_data[row[idx]] = d
                else:  # row is background, parse PC's
                    bg_pc1.append(float(row["PC1"]))
                    bg_pc2.append(float(row["PC2"]))
                    bg_ancestry.append(row["given_ancestry"])
                    # background pc's are added to parsed_data later

            # check that something was parsed:
            if len(parsed_data) > 0:
                # add background principal components
                self.somalier_background_pcs["background_pcs"] = {"PC1": bg_pc1, "PC2": bg_pc2, "ancestry": bg_ancestry}

                # cycle over keys, i.e. sample names
                # safely add new data to object data
                # warn when overwriting
                for s_name_raw in parsed_data:
                    s_name = "*".join([self.clean_s_name(s, f) for s in s_name_raw.split("*")])
                    if s_name in self.somalier_data.keys():
                        intersect_keys = parsed_data[s_name_raw].keys() & self.somalier_data.keys()
                        if len(intersect_keys) > 0:
                            log.debug(f"Duplicate sample name found! Overwriting: {s_name} : {intersect_keys}")
                    self.add_data_source(f, s_name)
                    try:
                        self.somalier_data[s_name].update(parsed_data[s_name_raw])
                    except KeyError:
                        self.somalier_data[s_name] = parsed_data[s_name_raw]
        else:
            log.warning(f"Detected empty file: {f['fn']}")

    def somalier_stats_table(self):
        """Add data to somalier stats table

        Bigger table within the somalier module, showing more stats"""

        headers = {
            "phenotype": {
                "title": "Phenotype",
                "description": "Sample's phenotype from pedigree info",
                "hidden": True,
            },
            "original_pedigree_sex": {
                "title": "Sex",
                "description": "Sample's sex from pedigree info",
                "scale": False,
            },
            "paternal_id": {
                "title": "Father ID",
                "description": "ID of sample's father ",
                "scale": False,
                "hidden": True,
            },
            "maternal_id": {
                "title": "Mother ID",
                "description": "ID of sample's mother",
                "scale": False,
                "hidden": True,
            },
            "family_id": {
                "title": "Family ID",
                "description": "ID of sample's family",
                "scale": False,
                "hidden": True,
            },
            "sex": {
                "title": "Inferred sex",
                "description": "Sample's inferred sex",
                "scale": False,
                "hidden": True,
            },
            "ancestry": {"title": "Ancestry", "description": "Most probable ancestry background", "scale": False},
            "p_ancestry": {
                "title": "P(Ancestry)",
                "description": "Ancestry probability",
                "max": 1,
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.2f}",
            },
            "n_het": {
                "title": "HetVar",
                "description": "Heterozygous variants",
                "shared_key": "variant_count",
                "format": "{:,.0f}",
            },
            "n_hom_ref": {
                "title": "HomRefVar",
                "description": "Homozygous reference variants",
                "shared_key": "variant_count",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "n_hom_alt": {
                "title": "HomAltVar",
                "description": "Homozygous alternate variants",
                "shared_key": "variant_count",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "n_unknown": {"title": "NA sites", "description": "Unknown sites", "format": "{:,.0f}"},
            "depth_mean": {
                "title": "Mean depth",
                "description": "Mean depth of all sites",
                "scale": "RdYlGn",
                "suffix": " X",
                "hidden": True,
            },
            "depth_sd": {
                "title": "Depth std",
                "description": "Depth's standard deviation of all sites",
                "scale": "RdYlGn",
                "hidden": True,
            },
            "gt_depth_mean": {
                "title": "Sites depth",
                "description": "Mean depth of genotyped sites",
                "scale": "RdYlGn",
                "suffix": " X",
            },
            "gt_depth_sd": {
                "title": "Genot depth std",
                "description": "Depth's standard deviation of genotype sites",
                "scale": "RdYlGn",
                "suffix": " X",
                "hidden": True,
            },
            "ab_mean": {
                "title": "Allele balance",
                "description": "Mean allele balance",
                "scale": "RdYlGn",
            },
            "ab_std": {
                "title": "Allele balance std",
                "description": "Standard deviation of allele balance",
                "scale": "RdYlGn",
                "hidden": True,
            },
            "p_middling_ab": {
                "title": "Allele balance < 0.2, > 0.8",
                "description": "Proportion of sites with allele balance < 0.2 or > 0.8",
                "max": 1,
                "min": 0,
                "scale": "RdYlGn",
                "format": "{:,.2f}",
            },
            "X_het": {
                "title": "HetVar X",
                "description": "Heterozygous variants on X chromosome",
                "shared_key": "variant_count_xy",
                "format": "{:,.0f}",
            },
            "X_hom_ref": {
                "title": "HomRefVar X",
                "description": "Homozygous reference variants on X chromosome",
                "shared_key": "variant_count_xy",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "X_hom_alt": {
                "title": "HomAltVar X",
                "description": "Homozygous alternate variants on X chromosome",
                "shared_key": "variant_count_xy",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "X_n": {
                "title": "Sites X",
                "description": "Total sites on X chromosome",
                "shared_key": "variant_count_xy",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "X_depth_mean": {
                "title": "Mean depth X",
                "description": "Mean depth of sites on X chromosome",
                "scale": "RdYlGn",
                "suffix": " X",
            },
            "Y_n": {
                "title": "Sites Y",
                "description": "Total sites on Y chromosome",
                "shared_key": "variant_count_xy",
                "format": "{:,.0f}",
                "hidden": True,
            },
            "Y_depth_mean": {
                "title": "Mean depth Y",
                "description": "Mean depth of sites on Y chromosome",
                "scale": "RdYlGn",
                "suffix": " X",
            },
        }

        t_config = {
            "id": "somalier_stats",
            "namespace": "Somalier",
            "title": "Somalier: Statistics",
            "no_violin": True,
            "raw_data_fn": "multiqc_somalier_stats",
        }

        self.add_section(
            name="Statistics",
            anchor="somalier-stats",
            description="Various statistics from the somalier report.",
            plot=table.plot(self.somalier_data, headers, t_config),
        )

    def somalier_relatedness_plot(self):
        alpha = 0.6
        relatedness_groups = {
            0: {
                "name": "Unrelated",
                "color": f"rgba(74, 124, 182, {alpha})",
            },
            0.49: {
                "name": "Sib-sib",
                "color": f"rgba(243, 123, 40, {alpha})",
            },
            0.5: {
                "name": "Parent-child",
                "color": f"rgba(159, 84, 47, {alpha})",
            },
        }

        # Get index colour scale
        cscale = mqc_colour.mqc_colour_scale()
        extra_colours = cscale.get_colours("Dark2")
        extra_colours = _make_col_alpha(extra_colours, alpha)
        extra_colour_idx = 0
        data = dict()
        for pair, d in self.somalier_data.items():
            if "expected_relatedness" not in d:
                continue

            relatedness = d["expected_relatedness"]
            # -1 is not the same family, 0 is same family but unrelated
            # @brentp says he usually bundles them together
            if relatedness == -1:
                relatedness = 0

            # New unique value that we've not seen before
            if relatedness not in relatedness_groups:
                relatedness_groups[relatedness] = {
                    "name": str(relatedness),
                    "color": extra_colours[extra_colour_idx],
                }
                extra_colour_idx += 0
                if extra_colour_idx > len(extra_colours):
                    extra_colour_idx = 0

            data[pair] = {
                "x": d["ibs0"],
                "y": d["ibs2"],
                "color": relatedness_groups[relatedness]["color"],
                "group": relatedness_groups[relatedness]["name"],
            }

        if len(data) == 0:
            return

        pconfig = {
            "id": "somalier_relatedness_plot",
            "title": "Somalier: Sample Shared Allele Rates (IBS)",
            "xlab": "IBS0 (no alleles shared)",
            "ylab": "IBS2 (both alleles shared)",
            "marker_line_width": 0,
        }

        colours_legend = ""
        for rel, group in sorted(relatedness_groups.items()):
            col = group["color"].replace(str(alpha), "1.0")
            colours_legend += f'<span style="color:{col}">{group["name"]}</span>, '

        self.add_section(
            name="Relatedness",
            anchor="somalier-relatedness",
            description=f"""
            Shared allele rates between sample pairs.
            Points are coloured by degree of expected relatedness: {colours_legend}""",
            plot=scatter.plot(data, pconfig),
        )

    def somalier_relatedness_heatmap_plot(self):
        # inspiration: MultiQC/modules/vcftools/relatedness2.py

        data = []
        labels = set()
        rels = defaultdict(dict)
        for s_name, d in self.somalier_data.items():
            if "relatedness" in d:
                a, b = s_name.split("*")
                labels.add(a)
                labels.add(b)
                rels[a][b] = rels[b][a] = d["relatedness"]
                rels[a][a] = rels[b][b] = 1.0

        # impose alphabetical order and avoid json serialisation errors in utils.report
        labels = sorted(labels)

        for x in labels:
            line = []
            for y in labels:
                try:
                    line.append(rels[x][y])
                except KeyError:
                    line.append(None)
            data.append(line)

        if len(data) > 0:
            pconfig = {
                "id": "somalier_relatedness_heatmap_plot",
                "title": "Somalier: Sample Relatedness",
                "xlab": "Sample A",
                "ylab": "Sample B",
            }

            self.add_section(
                name="Relatedness Heatmap",
                anchor="somalier-relatedness-heatmap",
                description="Heatmap displaying relatedness of sample pairs.",
                plot=heatmap.plot(
                    data=data,
                    xcats=labels,
                    ycats=labels,
                    pconfig=pconfig,
                ),
            )

    def somalier_het_check_plot(self):
        """plot the het_check scatter plot"""
        # empty dictionary to add sample names, and dictionary of values
        data = {}

        # for each sample, and list in self.somalier_data
        for s_name, d in self.somalier_data.items():
            # check the sample contains the required columns
            if "gt_depth_mean" in d and "ab_std" in d:
                # add sample to dictionary with value as a dictionary of points to plot
                data[s_name] = {"x": d["gt_depth_mean"], "y": d["ab_std"]}

        if len(data) > 0:
            pconfig = {
                "id": "somalier_het_check_plot",
                "title": "Somalier: Sample Observed Heterozygosity",
                "xlab": "Mean depth",
                "ylab": "Standard deviation of allele-balance",
            }

            self.add_section(
                name="Heterozygosity",
                description="Standard deviation of heterozygous allele balance against mean depth.",
                helptext="A high standard deviation in allele balance suggests contamination.",
                anchor="somalier-hetcheck",
                plot=scatter.plot(data, pconfig),
            )

    def somalier_sex_check_plot(self):
        data = {}
        sex_index = {"female": 0, "male": 1, "unknown": 2}

        for s_name, d in self.somalier_data.items():
            if "X_depth_mean" in d and "original_pedigree_sex" in d:
                if d["gt_depth_mean"] == 0:
                    y = 0
                else:
                    y = 2 * d["X_depth_mean"] / d["gt_depth_mean"]
                data[s_name] = {
                    "x": sex_index.get(d["original_pedigree_sex"], 2) + (random.random() - 0.5) * 0.1,
                    "y": y,
                }

        if len(data) > 0:
            pconfig = {
                "id": "somalier_sex_check_plot",
                "title": "Somalier: Sample Predicted Sex",
                "xlab": "Sex from pedigree",
                "ylab": "Scaled mean depth on X",
                "categories": ["Female", "Male", "Unknown"],
            }

            self.add_section(
                name="Sex",
                description="Predicted sex against scaled depth on X",
                helptext="Higher values of depth, low values suggest male.",
                anchor="somalier-sexcheck",
                plot=scatter.plot(data, pconfig),
            )

    def somalier_ancestry_barplot(self):
        data = dict()
        c_scale = mqc_colour.mqc_colour_scale(name="Paired").colours
        cats = dict()
        anc_cats = self.somalier_ancestry_cats

        # use Paired color scale, unless number of categories exceed colors
        if len(anc_cats) <= len(c_scale):
            for i in range(len(anc_cats)):
                c = anc_cats[i]
                if i < (len(c_scale) - 1):
                    col = c_scale[i]
                else:
                    # default col if more cats than cols
                    col = "rgb(211,211,211,0.5)"
                cats[c] = {"name": c, "color": col}
        else:
            cats = None

        for s_name, d in self.somalier_data.items():
            # ensure that only relevant items are added,
            # i.e. only ancestry category values
            ls = {k: v * 100.0 for k, v in d.items() if (k in self.somalier_ancestry_cats)}
            if len(ls) > 0:  # only add dict, if it contains values
                data[s_name] = ls

        if len(data) > 0:
            pconfig = {
                "id": "somalier_ancestry_barplot",
                "title": "Somalier: Sample Predicted Ancestry Proportions",
                "cpswitch": False,
                "hide_zero_cats": False,
                "ylab": "Predicted Ancestry",
                "tt_suffix": "%",
            }

            self.add_section(
                name="Ancestry Barplot",
                description="Predicted ancestries of samples.",
                helptext="""
                Shows the percentwise predicted probability of each
                ancestry. A sample might contain traces of several ancestries.
                If the number of samples is too high, the plot is rendered as a
                non-interactive flat image.
                """,
                anchor="somalier-ancestry",
                plot=bargraph.plot(data=data, cats=cats, pconfig=pconfig),
            )

    def somalier_ancestry_pca_plot(self):
        data = dict()

        # add background
        # N.B. this must be done after samples to have samples on top
        d = self.somalier_background_pcs.pop("background_pcs", {})
        if d:
            # generate color scale to match the number of categories
            c_scale = mqc_colour.mqc_colour_scale(name="Paired").colours
            cats = self.somalier_ancestry_cats
            ancestry_colors = dict(zip(cats, c_scale[: len(cats)]))
            default_background_color = "rgba(255,192,203,0.3)"

            # Make colours semi-transparent
            ancestry_colors = dict(zip(ancestry_colors.keys(), _make_col_alpha(ancestry_colors.values(), 0.3)))

            background = [
                {
                    "x": pc1,
                    "y": pc2,
                    "color": ancestry_colors.get(ancestry, default_background_color),
                    "name": ancestry,
                    "marker_size": 3,
                    "marker_line_width": 0,
                    "annotate": False,
                }
                for pc1, pc2, ancestry in zip(d["PC1"], d["PC2"], d["ancestry"])
            ]
            data["background"] = background

        # cycle over samples and add PC coordinates to data dict
        for s_name, d in self.somalier_data.items():
            if "PC1" in d and "PC2" in d:
                data[s_name] = {
                    "x": d["PC1"],
                    "y": d["PC2"],
                    "color": "rgba(0, 0, 0, 0.6)",
                }

        # generate section and plot
        if len(data) > 0:
            pconfig = {
                "id": "somalier_ancestry_pca_plot",
                "title": "Somalier: Sample Predicted Ancestry",
                "xlab": "PC1",
                "ylab": "PC2",
                "marker_size": 5,
            }

            self.add_section(
                name="Ancestry PCA",
                description="Principal components of samples against background PCs.",
                helptext="""
                Sample PCs are plotted against background PCs from the
                background data supplied to somalier.
                Color indicates predicted ancestry of sample. Data points in close
                proximity are predicted to be of similar ancestry. Consider whether
                the samples cluster as expected.
                """,
                anchor="somalier-ancestry-pca",
                plot=scatter.plot(data, pconfig),
            )


def _make_col_alpha(cols, alpha):
    """Take an HTML colour value and return a rgba string with alpha"""
    cols_return = []
    for col in cols:
        col_srgb = spectra.html(col)
        cols_rgb = [c * 255.0 for c in col_srgb.clamped_rgb]
        cols_return.append("rgba({},{},{},{})".format(*cols_rgb, alpha))
    return cols_return



"""
=========
 default
=========

The main MultiQC report template, lovingly known to its admirers
simply as "default"

Note, this is where most of the MultiQC report interactive functionality
is based and will be developed. Unless you want to do some really radical
changes, you probably don't want to replace this theme. Instead, you can
create a child theme that starts with 'default' and then overwrites
certain files.

For more information about creating child themes, see the docs:
docs/templates.md

"""
import os

template_dir = os.path.dirname(__file__)
base_fn = "base.html"



"""
=========
 gathered
=========

This theme extends 'default'.  Unlike 'default', it organizes all sections
under the 'Detailed Statistics' header, rather than grouping them by 
module.
"""
import os

template_dir = os.path.dirname(__file__)
template_parent = "default"
base_fn = "base.html"



""" MultiQC module to parse output from HTSeq Count """


import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="HTSeq Count",
            anchor="htseq",
            target="HTSeq Count",
            href="https://htseq.readthedocs.io/en/master/htseqcount.html",
            info=" is part of the HTSeq Python package - it takes a file with aligned sequencing "
            "reads, plus a list of genomic features and counts how many reads map to each feature.",
            doi="10.1093/bioinformatics/btu638",
        )

        # Find and load any HTSeq Count reports
        self.htseq_data = dict()
        self.htseq_keys = list()
        for f in self.find_log_files("htseq", filehandles=True):
            parsed_data = self.parse_htseq_report(f)
            if parsed_data is not None:
                self.htseq_data[f["s_name"]] = parsed_data
                self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.htseq_data = self.ignore_samples(self.htseq_data)

        if len(self.htseq_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.htseq_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.htseq_data, "multiqc_htseq")

        # Basic Stats Table
        self.htseq_stats_table()

        # Assignment bar plot
        self.add_section(plot=self.htseq_counts_chart())

    @staticmethod
    def parse_htseq_report(f):
        """Parse the HTSeq Count log file."""
        keys = ["__no_feature", "__ambiguous", "__too_low_aQual", "__not_aligned", "__alignment_not_unique"]
        parsed_data = dict()
        assigned_counts = 0

        # HtSeq search pattern is just two tab-separated columns, which is not very specific.
        # Need to wrap in try-catch to catch potential weird files like parquet being matched.
        try:
            for line in f["f"]:
                s = line.split("\t")
                if s[0] in keys:
                    parsed_data[s[0][2:]] = int(s[-1])
                else:
                    try:
                        assigned_counts += int(s[-1])
                    except (ValueError, IndexError):
                        pass
        except UnicodeDecodeError:
            log.debug(f"Could not parse potential HTSeq Count file {f['fn']}")
            return None

        if len(parsed_data) > 0:
            parsed_data["assigned"] = assigned_counts
            parsed_data["total_count"] = sum([v for v in parsed_data.values()])
            try:
                parsed_data["percent_assigned"] = (
                    float(parsed_data["assigned"]) / float(parsed_data["total_count"])
                ) * 100.0
            except ZeroDivisionError:
                parsed_data["percent_assigned"] = 0
            return parsed_data
        return None

    def htseq_stats_table(self):
        """Take the parsed stats from the HTSeq Count report and add them to the
        basic stats table at the top of the report"""

        headers = {
            "percent_assigned": {
                "title": "% Assigned",
                "description": "% Assigned reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "assigned": {
                "title": f"{config.read_count_prefix} Assigned",
                "description": f"Assigned Reads ({config.read_count_desc})",
                "min": 0,
                "scale": "PuBu",
                "modify": lambda x: float(x) * config.read_count_multiplier,
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.htseq_data, headers)

    def htseq_counts_chart(self):
        """Make the HTSeq Count assignment rates plot"""
        cats = {
            "assigned": {"name": "Assigned"},
            "ambiguous": {"name": "Ambiguous"},
            "alignment_not_unique": {"name": "Alignment Not Unique"},
            "no_feature": {"name": "No Feature"},
            "too_low_aQual": {"name": "Too Low aQual"},
            "not_aligned": {"name": "Not Aligned"},
        }
        config = {
            "id": "htseq_assignment_plot",
            "title": "HTSeq: Count Assignments",
            "ylab": "# Reads",
            "hide_zero_cats": False,
            "cpswitch_counts_label": "Number of Reads",
        }
        return bargraph.plot(self.htseq_data, cats, config)



""" MultiQC module to parse output from Nextclade """

import csv
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Nextclade",
            anchor="nextclade",
            href="https://github.com/nextstrain/nextclade",
            info="does viral genome alignment, clade assignment, mutation calling, and quality checks",
            doi="10.21105/joss.03773",
        )

        # Parse logs
        self.nextclade_data = dict()
        for f in self.find_log_files("nextclade", filehandles=True):
            self.parse_nextclade_log(f)
            self.add_data_source(f)

        # Remove ignored samples
        self.nextclade_data = self.ignore_samples(self.nextclade_data)

        # Check if we found any samples
        if len(self.nextclade_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.nextclade_data)} samples")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.nextclade_data, "multiqc_nextclade")

        # Add data to general statistics table
        self.nextclade_general_stats_table()

        # Create the Nextclade table
        self.add_section(
            name="Run table", anchor="nextclade-run", description="", helptext="", plot=self.nextclade_table()
        )

    def parse_nextclade_log(self, f):
        for sample in csv.DictReader(f["f"], delimiter=";"):
            try:
                s_name = sample.pop("seqName")
                s_name = self.clean_s_name(s_name, f)
                if s_name in self.nextclade_data:
                    log.error(f"Found duplicate sample '{s_name}' - overwriting")

                # Clean column names: make lower case and
                # replace '.' with '_'
                for key in list(sample.keys()):
                    new_key = key.lower()
                    if "." in new_key:
                        new_key = "_".join(new_key.split("."))
                    sample[new_key] = sample.pop(key)

                for key in list(sample.keys()):
                    new_key = key.lower()
                    sample[new_key] = sample.pop(key)
                    try:
                        sample[new_key] = float(sample[new_key])
                    except ValueError:
                        pass

                self.nextclade_data[s_name] = sample
            except KeyError:
                log.error(f"File '{f['fn']}' could not be parsed - no sequence name found")

    def nextclade_general_stats_table(self):
        """Takes the parsed sample data and adds the sequences' clade to the general stats table"""

        headers = {
            "clade": {
                "title": "Clade",
                "description": "Clade",
                "scale": False,
            }
        }
        self.general_stats_addcols(self.nextclade_data, headers)

    def nextclade_table(self):
        """Creates the table of data for the samples"""

        headers = {
            "clade": {
                "title": "Clade",
                "description": "The inferred clade from the input sequence and reference tree",
                "scale": False,
            },
            "qc_overallscore": {
                "title": "QC Overall Score",
                "description": "Summarizing score of all QC test",
                "hidden": True,
            },
            "qc_overallstatus": {
                "title": "QC Overall Status",
                "description": "Summarizing status of all QC tests",
                "scale": False,
                "cond_formatting_rules": {
                    "pass": [{"s_eq": "good"}],
                    "fail": [{"s_eq": "bad"}],
                },
            },
            "totalgaps": {
                "title": "Total Gaps",
                "description": "Total number of detected nucleotide gaps",
                "min": 0,
                "scale": "RdBu-rev",
                "hidden": True,
            },
            "totalinsertions": {
                "title": "Total Insertions",
                "description": "Total number of detected nucleotide insertions",
                "min": 0,
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "totalmissing": {
                "title": "Total Missing",
                "description": "Total number of detected missing nucleotides",
                "min": 0,
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "totalmutations": {
                "title": "Total Mutations",
                "description": "Total number of detected nucleotide mutations",
                "min": 0,
                "scale": "RdBu-rev",
                "hidden": True,
            },
            "totalnonacgtns": {
                "title": "Total Non-ACGTNs",
                "description": "Total number of detected ambiguous nucleotides",
                "min": 0,
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "totalpcrprimerchanges": {
                "title": "Total PCR Primer Changes",
                "description": "Total number of nucleotide mutations detected in PCR primer regions",
                "min": 0,
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "totalaminoacidsubstitutions": {
                "title": "Total Amino Acid Substitutions",
                "description": "Total number of detected aminoacid substitutions",
                "min": 0,
                "scale": "RdBu-rev",
                "hidden": True,
            },
            "totalaminoaciddeletions": {
                "title": "Total Amino Acid Deletions",
                "description": "Total number of detected aminoacid substitutions",
                "min": 0,
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "alignmentscore": {
                "title": "Alignment Score",
                "description": "Indicates to what degree the input sequence and the reference sequence correspond",
                "min": 0,
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "alignmentstart": {
                "title": "Alignment Start",
                "description": "Beginning of the sequenced region",
                "hidden": True,
            },
            "alignmentend": {
                "title": "Alignment End",
                "description": "End of the sequenced region",
                "hidden": True,
            },
            "qc_missingdata_missingdatathreshold": {
                "title": "QC Missing Data Threshold",
                "description": "The threshold used for the 'Missing data' QC test",
                "scale": "GnBu",
                "hidden": True,
            },
            "qc_missingdata_score": {
                "title": "QC Missing Data Score",
                "description": "Score from the 'Missing data' QC test",
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "qc_missingdata_status": {
                "title": "QC Missing Data Status",
                "description": "Status for 'Missing data' QC test",
                "scale": False,
                "cond_formatting_rules": {
                    "pass": [{"s_eq": "good"}],
                    "fail": [{"s_eq": "bad"}],
                },
            },
            "qc_missingdata_totalmissing": {
                "title": "QC Missing Data Total",
                "description": "Total number of missing nucleotides used in 'Missing data' QC test",
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "qc_mixedsites_mixedsitesthreshold": {
                "title": "QC Mixed Sites Threshold",
                "description": "Threshold used for 'Mixed sites' QC test",
                "scale": "GnBu",
                "hidden": True,
            },
            "qc_mixedsites_score": {
                "title": "QC Mixed Sites Score",
                "description": "Score from the 'Missing data' QC testScore from the 'Missing data' QC test",
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "qc_mixedsites_status": {
                "title": "QC Mixed Sites Status",
                "description": "Status for 'Missing data' QC test",
                "scale": False,
                "cond_formatting_rules": {
                    "pass": [{"s_eq": "good"}],
                    "fail": [{"s_eq": "bad"}],
                },
            },
            "qc_mixedsites_totalmixedsites": {
                "title": "QC Mixed Sites Total",
                "description": "Total number of ambiguous nucleotides used for 'Mixed sites' QC test",
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "qc_privatemutations_cutoff": {
                "title": "QC Private Mutations Cutoff",
                "description": "Cutoff parameter used for 'Private mutations' QC test",
                "scale": "GnBu",
                "hidden": True,
            },
            "qc_privatemutations_excess": {
                "title": "QC Private Mutations Excess",
                "description": "Excess parameter used for 'Private mutations' QC test",
                "scale": "OrRd",
                "hidden": True,
            },
            "qc_privatemutations_score": {
                "title": "QC Private Mutations Score",
                "description": "Score for 'Private mutations' QC rule",
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "qc_privatemutations_status": {
                "title": "QC Private Mutations Status",
                "description": "Status for 'Private mutations' QC rule",
                "scale": False,
                "cond_formatting_rules": {
                    "pass": [{"s_eq": "good"}],
                    "fail": [{"s_eq": "bad"}],
                },
                "hidden": True,
            },
            "qc_privatemutations_total": {
                "title": "QC Private Mutations Total",
                "description": "Total number of private mutations used for 'Private mutations' QC rule",
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
            "qc_snpclusters_clusteredsnps": {
                "title": "QC Clustered SNPs",
                "description": "Clustered SNP detected for 'SNP clusters' QC test",
                "hidden": True,
            },
            "qc_snpclusters_score": {
                "title": "QC SNP Clusters Score",
                "description": "Score for 'SNP clusters' QC test",
                "scale": "RdYlGn-rev",
                "hidden": True,
            },
            "qc_snpclusters_status": {
                "title": "QC SNP Clusters Status",
                "description": "Status for 'SNP clusters' QC test",
                "scale": False,
                "cond_formatting_rules": {
                    "pass": [{"s_eq": "good"}],
                    "fail": [{"s_eq": "bad"}],
                },
                "hidden": True,
            },
            "qc_snpclusters_totalsnps": {
                "title": "QC Total SNPs",
                "description": "Total number of SNPs for 'SNP clusters' QC test",
                "scale": "RdYlBu-rev",
                "hidden": True,
            },
        }

        # Main table config
        table_config = {
            "namespace": "Nextclade",
            "id": "nextclade_run_table",
            "table_title": "Nextclade Run details",
        }

        return table.plot(self.nextclade_data, headers, table_config)



""" MultiQC module to parse output from sickle """

import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Sickle",
            anchor="sickle",
            href="https://github.com/najoshi/sickle",
            info="A windowed adaptive trimming tool for FASTQ files using quality.",
            # No DOI // doi=
        )

        # parse list of log files
        self.sickle_data = dict()
        for f in self.find_log_files("sickle"):
            parsed_data = self.parse_logs(f["f"])
            if len(parsed_data):
                if f["s_name"] in self.sickle_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.sickle_data[f["s_name"]] = parsed_data
                self.add_data_source(f)

        self.sickle_data = self.ignore_samples(self.sickle_data)

        # no file found
        if len(self.sickle_data) == 0:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.sickle_data, "multiqc_sickle")

        self.sickle_general_stats_table()
        self.read_count_plot()

    @staticmethod
    def parse_logs(f):
        """Parse the Sickle standard output"""
        regexes = [
            # Paired-end
            ["reads_paired_kept", re.compile(r"FastQ paired records kept: ([\d,]+) .*")],
            ["reads_single_kept", re.compile(r"FastQ single records kept: ([\d,]+).*")],
            ["reads_paired_discarded", re.compile(r"FastQ paired records discarded: ([\d,]+) .*")],
            ["reads_single_discarded", re.compile(r"FastQ single records discarded: ([\d,]+) .*")],
            # Single-end
            ["reads_single_kept", re.compile(r"FastQ records kept: ([\d,]+)")],
            ["reads_single_discarded", re.compile(r"FastQ records discarded: ([\d,]+)")],
        ]
        data = {}
        for line in f.splitlines():
            # Search regexes for overview stats
            for k, regex in regexes:
                match = regex.search(line)
                if match:
                    data[k] = int(match.group(1).replace(",", ""))

        # if paired data : compute total for general stats table
        if "reads_single_kept" in data:
            data["reads_total_kept"] = data.get("reads_paired_kept", 0) + data["reads_single_kept"]

        if "reads_single_discarded" in data:
            data["reads_total_discarded"] = data.get("reads_paired_discarded", 0) + data["reads_single_discarded"]

        # Calculate percentage discarded
        if "reads_total_kept" in data and "reads_single_discarded" in data:
            data["percentage_discarded"] = (
                float(data["reads_total_discarded"]) / float(data["reads_total_kept"] + data["reads_total_discarded"])
            ) * 100.0

        return data

    def sickle_general_stats_table(self):
        """Take the parsed stats from the sickle report and add
        number of kept reads into the generam stats table"""

        headers = {
            "percentage_discarded": {
                "title": "% Reads discarded",
                "description": "Percentage of reads discarded",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            },
            "reads_total_kept": {
                "title": f"{config.read_count_prefix} Reads kept",
                "description": f"Number of reads kept ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
                "scale": "BuGn",
            },
        }
        self.general_stats_addcols(self.sickle_data, headers)

    def read_count_plot(self):
        """Stacked bar plot showing counts of reads"""
        pconfig = {
            "id": "sickle_sequence_counts_plot",
            "title": "Sickle: Sequence Counts",
            "ylab": "Number of reads",
            "cpswitch_counts_label": "Number of reads",
        }
        pcats = {
            "reads_paired_kept": {"name": "Paired reads kept", "color": "#1f78b4"},
            "reads_single_kept": {"name": "Single reads kept", "color": "#a6cee3"},
            "reads_paired_discarded": {"name": "Paired reads discarded", "color": "#ff7f00"},
            "reads_single_discarded": {"name": "Single reads discarded", "color": "#fdae61"},
        }

        self.add_section(
            name="Sequence counts",
            anchor="sickle_sequence_counts",
            description="Number of reads discarded by Sickle.",
            plot=bargraph.plot(self.sickle_data, pcats, pconfig),
        )



""" MultiQC module to parse output from Samtools rmdup """

import logging
import re

from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class RmdupReportMixin:
    def parse_samtools_rmdup(self):
        """Find Samtools rmdup logs and parse their data"""

        self.samtools_rmdup = dict()
        for f in self.find_log_files("samtools/rmdup", filehandles=True):
            # Example below:
            # [bam_rmdupse_core] 26602816 / 103563641 = 0.2569 in library '   '
            dups_regex = r"\[bam_rmdups?e?_core\] (\d+) / (\d+) = (\d+\.\d+) in library '(.*)'"
            s_name = f["s_name"]
            for line in f["f"]:
                match = re.search(dups_regex, line)
                if match:
                    library_name = match.group(4).strip()
                    if library_name != "":
                        s_name = library_name
                    if s_name in self.samtools_rmdup:
                        log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.samtools_rmdup[s_name] = {}
                    self.samtools_rmdup[s_name]["n_dups"] = int(match.group(1))
                    self.samtools_rmdup[s_name]["n_tot"] = int(match.group(2))
                    self.samtools_rmdup[s_name]["n_unique"] = int(match.group(2)) - int(match.group(1))
                    self.samtools_rmdup[s_name]["pct_dups"] = float(match.group(3)) * 100

        # Filter to strip out ignored sample names
        self.samtools_rmdup = self.ignore_samples(self.samtools_rmdup)

        if len(self.samtools_rmdup) == 0:
            return 0

        # Write parsed report data to a file
        self.write_data_file(self.samtools_rmdup, "multiqc_samtools_rmdup")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Make a bar plot showing duplicates
        keys = {
            "n_unique": {"name": "Non-duplicated reads"},
            "n_dups": {"name": "Duplicated reads"},
        }
        pconfig = {
            "id": "samtools_rmdup_plot",
            "title": "Samtools rmdup: Duplicate alignments",
            "ylab": "Number of reads",
            "yDecimals": False,
        }
        self.add_section(
            name="Duplicates removed",
            anchor="samtools-rmdup",
            plot=bargraph.plot(self.samtools_rmdup, keys, pconfig),
        )

        # Add a column to the General Stats table
        # General Stats Table
        stats_headers = {
            "pct_dups": {
                "title": "Duplicates",
                "description": "Percent of duplicate alignments",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
            }
        }
        self.general_stats_addcols(self.samtools_rmdup, stats_headers, namespace="rmdup")

        return len(self.samtools_rmdup)



""" MultiQC submodule to parse output from RSeQC bam_stat.py
http://rseqc.sourceforge.net/#bam-stat-py """

import logging
import re

from multiqc.plots import beeswarm
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(self):
    """Find RSeQC bam_stat reports and parse their data"""

    # Set up vars
    self.bam_stat_data = dict()
    regexes = {
        "total_records": r"Total records:\s*(\d+)",
        "qc_failed": r"QC failed:\s*(\d+)",
        "optical_pcr_duplicate": r"Optical/PCR duplicate:\s*(\d+)",
        "non_primary_hits": r"Non primary hits\s*(\d+)",
        "unmapped_reads": r"Unmapped reads:\s*(\d+)",
        "mapq_lt_mapq_cut_non-unique": r"mapq < mapq_cut \(non-unique\):\s*(\d+)",
        "mapq_gte_mapq_cut_unique": r"mapq >= mapq_cut \(unique\):\s*(\d+)",
        "read_1": r"Read-1:\s*(\d+)",
        "read_2": r"Read-2:\s*(\d+)",
        "reads_map_to_sense": r"Reads map to '\+':\s*(\d+)",
        "reads_map_to_antisense": r"Reads map to '-':\s*(\d+)",
        "non-splice_reads": r"Non-splice reads:\s*(\d+)",
        "splice_reads": r"Splice reads:\s*(\d+)",
        "reads_mapped_in_proper_pairs": r"Reads mapped in proper pairs:\s*(\d+)",
        "proper-paired_reads_map_to_different_chrom": r"Proper-paired reads map to different chrom:\s*(\d+)",
    }

    # intiate PE check
    is_paired_end = False

    # Go through files and parse data using regexes
    for f in self.find_log_files("rseqc/bam_stat"):
        d = dict()
        for k, r in regexes.items():
            r_search = re.search(r, f["f"], re.MULTILINE)
            if r_search:
                d[k] = int(r_search.group(1))

        # Calculate some percentages
        if "total_records" in d:
            t = float(d["total_records"])
            try:
                d["unique_percent"] = (float(d["mapq_gte_mapq_cut_unique"]) / t) * 100.0
            except (KeyError, ZeroDivisionError):
                pass
            try:
                d["proper_pairs_percent"] = (float(d["reads_mapped_in_proper_pairs"]) / t) * 100.0
            except (KeyError, ZeroDivisionError):
                pass

        if len(d) > 0:
            if f["s_name"] in self.bam_stat_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f, section="bam_stat")
            # Check if SE or PE
            if d["read_2"] != 0:
                is_paired_end = True
            self.bam_stat_data[f["s_name"]] = d

    # Filter to strip out ignored sample names
    self.bam_stat_data = self.ignore_samples(self.bam_stat_data)

    if len(self.bam_stat_data) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Write to file
    self.write_data_file(self.bam_stat_data, "multiqc_rseqc_bam_stat")

    # Add to general stats table
    self.general_stats_headers["proper_pairs_percent"] = {
        "title": "% Proper Pairs",
        "description": "% Reads mapped in proper pairs",
        "max": 100,
        "min": 0,
        "suffix": "%",
        "scale": "RdYlGn",
    }
    for s_name in self.bam_stat_data:
        if s_name not in self.general_stats_data:
            self.general_stats_data[s_name] = dict()

            # Only write if PE, i.e. there is something to write
            if is_paired_end:
                self.general_stats_data[s_name].update(self.bam_stat_data[s_name])

    # Make dot plot of counts
    pconfig = {
        "id": "rseqc_bam_stat",
        "title": "RSeQC: Bam Stat",
    }
    defaults = {
        "min": 0,
        "shared_key": "read_count",
        "decimalPlaces": 2,
        "modify": lambda x: x * config.read_count_multiplier,
        "suffix": config.read_count_prefix,
    }
    keys = dict()
    keys["total_records"] = dict(defaults, **{"title": "Total records"})
    keys["qc_failed"] = dict(defaults, **{"title": "QC failed"})
    keys["optical_pcr_duplicate"] = dict(defaults, **{"title": "Duplicates", "description": "Optical/PCR duplicate"})
    keys["non_primary_hits"] = dict(defaults, **{"title": "Non primary hit"})
    keys["unmapped_reads"] = dict(defaults, **{"title": "Unmapped", "description": "Unmapped reads"})
    keys["mapq_lt_mapq_cut_non"] = dict(
        defaults, **{"title": "Non-unique", "description": "mapq < mapq_cut (non-unique)"}
    )
    keys["mapq_gte_mapq_cut_unique"] = dict(defaults, **{"title": "Unique", "description": "mapq >= mapq_cut (unique)"})
    if is_paired_end:
        keys["read_1"] = dict(defaults, **{"title": "Read-1"})
        keys["read_2"] = dict(defaults, **{"title": "Read-2"})
    keys["reads_map_to_sense"] = dict(defaults, **{"title": "+ve strand", "description": "Reads map to '+'"})
    keys["reads_map_to_antisense"] = dict(defaults, **{"title": "-ve strand", "description": "Reads map to '-'"})
    keys["non-splice_reads"] = dict(defaults, **{"title": "Non-splice reads"})
    keys["splice_reads"] = dict(defaults, **{"title": "Splice reads"})
    if is_paired_end:
        keys["reads_mapped_in_proper_pairs"] = dict(
            defaults, **{"title": "Proper pairs", "description": "Reads mapped in proper pairs"}
        )
        keys["proper-paired_reads_map_to_different_chrom"] = dict(
            defaults, **{"title": "Different chrom", "description": "Proper-paired reads map to different chrom"}
        )

    self.add_section(
        name="Bam Stat",
        anchor="rseqc-bam_stat",
        description="All numbers reported in millions.",
        plot=beeswarm.plot(self.bam_stat_data, keys, pconfig),
    )

    # Return number of samples found
    return len(self.bam_stat_data)



# -*- coding: utf-8 -*-

""" MultiQC submodule to parse output from Samtools stats """

import logging
import re

from multiqc import config
from multiqc.plots import bargraph, beeswarm

# Initialise the logger
log = logging.getLogger(__name__)

# Regex to grab version number from samtools stats contents
VERSION_REGEX = r"# This file was produced by samtools stats \(([\d\.]+)"
HTSLIB_REGEX = r"\+htslib-([\d\.]+)"


class StatsReportMixin:
    """Mixin class, loaded by main samtools MultiqcModule class."""

    def parse_samtools_stats(self):
        """Find Samtools stats logs and parse their data"""

        self.samtools_stats = dict()
        for f in self.find_log_files("samtools/stats"):
            parsed_data = dict()
            for line in f["f"].splitlines():
                # Get version number from file contents
                if line.startswith("# This file was produced by samtools stats"):
                    # Look for Samtools version
                    version_match = re.search(VERSION_REGEX, line)
                    if version_match is None:
                        continue

                    # Add Samtools version
                    samtools_version = version_match.group(1)
                    self.add_software_version(samtools_version, f["s_name"])

                    # Look for HTSlib version
                    htslib_version_match = re.search(HTSLIB_REGEX, line)
                    if htslib_version_match is None:
                        continue

                    # Add HTSlib version if different from Samtools version
                    htslib_version = htslib_version_match.group(1)
                    if htslib_version != samtools_version:
                        self.add_software_version(htslib_version, f["s_name"], "HTSlib")

                if not line.startswith("SN"):
                    continue
                sections = line.split("\t")
                field = sections[1].strip()[:-1]
                field = field.replace(" ", "_")
                value = float(sections[2].strip())
                parsed_data[field] = value

            if len(parsed_data) > 0:
                # Work out some percentages
                if "raw_total_sequences" in parsed_data:
                    for k in list(parsed_data.keys()):
                        if (
                            k.startswith("reads_")
                            and k != "raw_total_sequences"
                            and parsed_data["raw_total_sequences"] > 0
                        ):
                            parsed_data[f"{k}_percent"] = (parsed_data[k] / parsed_data["raw_total_sequences"]) * 100

                if f["s_name"] in self.samtools_stats:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f, section="stats")
                self.samtools_stats[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.samtools_stats = self.ignore_samples(self.samtools_stats)

        if len(self.samtools_stats) == 0:
            return 0

        # Write parsed report data to a file
        self.write_data_file(self.samtools_stats, "multiqc_samtools_stats")

        # General Stats Table
        stats_headers = {
            "error_rate": {
                "title": "Error rate",
                "description": "Error rate: mismatches (NM) / bases mapped (CIGAR)",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "OrRd",
                "format": "{:,.2f}",
                "modify": lambda x: x * 100.0,
            },
            "non-primary_alignments": {
                "title": "Non-primary",
                "description": f"Non-primary alignments ({config.read_count_desc})",
                "scale": "PuBu",
                "shared_key": "read_count",
            },
            "reads_mapped": {
                "title": "Reads mapped",
                "description": f"Reads mapped in the bam file ({config.read_count_desc})",
                "shared_key": "read_count",
            },
            "reads_mapped_percent": {
                "title": "% Mapped",
                "description": "% Mapped reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "reads_properly_paired_percent": {
                "title": "% Proper pairs",
                "description": "% Properly paired reads",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "RdYlGn",
                "hidden": True
                if (max([x["reads_mapped_and_paired"] for x in self.samtools_stats.values()]) == 0)
                else False,
            },
            "reads_MQ0_percent": {
                "title": "% MapQ 0 reads",
                "description": "% of reads that are ambiguously placed (MapQ=0)",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "OrRd",
                "hidden": True,
            },
            "raw_total_sequences": {
                "title": "Total seqs",
                "description": f"Total sequences in the bam file ({config.read_count_desc})",
                "shared_key": "read_count",
            },
        }
        self.general_stats_addcols(self.samtools_stats, stats_headers, namespace="stats")

        # Make bargraph plot of mapped/unmapped reads
        self.alignment_section(self.samtools_stats)

        # Make dot plot of counts
        keys = {}
        reads = {
            "min": 0,
            "modify": lambda x: float(x) * config.read_count_multiplier,
            "suffix": config.read_count_prefix,
            "decimalPlaces": 2,
            "shared_key": "read_count",
        }
        bases = {
            "min": 0,
            "modify": lambda x: float(x) * config.base_count_multiplier,
            "suffix": config.base_count_prefix,
            "decimalPlaces": 2,
            "shared_key": "base_count",
        }
        keys["raw_total_sequences"] = dict(reads, **{"title": "Total sequences"})
        keys["reads_mapped_and_paired"] = dict(
            reads,
            **{"title": "Mapped &amp; paired", "description": "Paired-end technology bit set + both mates mapped"},
        )
        keys["reads_properly_paired"] = dict(
            reads, **{"title": "Properly paired", "description": "Proper-pair bit set"}
        )
        keys["reads_duplicated"] = dict(
            reads, **{"title": "Duplicated", "description": "PCR or optical duplicate bit set"}
        )
        keys["reads_QC_failed"] = dict(reads, **{"title": "QC Failed"})
        keys["reads_MQ0"] = dict(reads, **{"title": "Reads MQ0", "description": "Reads mapped and MQ=0"})
        keys["bases_mapped_(cigar)"] = dict(
            bases, **{"title": "Mapped bases (CIGAR)", "description": "Mapped bases (CIGAR)"}
        )
        keys["bases_trimmed"] = dict(bases, **{"title": "Bases Trimmed"})
        keys["bases_duplicated"] = dict(bases, **{"title": "Duplicated bases"})
        keys["pairs_on_different_chromosomes"] = dict(
            reads, **{"title": "Diff chromosomes", "description": "Pairs on different chromosomes"}
        )
        keys["pairs_with_other_orientation"] = dict(
            reads, **{"title": "Other orientation", "description": "Pairs with other orientation"}
        )
        keys["inward_oriented_pairs"] = dict(reads, **{"title": "Inward pairs", "description": "Inward oriented pairs"})
        keys["outward_oriented_pairs"] = dict(
            reads, **{"title": "Outward pairs", "description": "Outward oriented pairs"}
        )

        self.add_section(
            name="Alignment stats",
            anchor="samtools-stats",
            description="This module parses the output from <code>samtools stats</code>. All numbers in millions.",
            plot=beeswarm.plot(
                self.samtools_stats,
                keys,
                {
                    "id": "samtools-stats-dp",
                    "title": "Samtools stats: Alignment Stats",
                },
            ),
        )

        # Return the number of logs that were found
        return len(self.samtools_stats)

    def alignment_section(self, samples_data):
        bedgraph_data = {}
        for sample_id, data in samples_data.items():
            # Breaking up the mapped reads count into MQ0 and >MQ1 counts
            data["reads_mapped_MQ1"] = data["reads_mapped"] - data["reads_MQ0"]
            # Asserting the bar plot keys sum up to the total
            expected_total = data["raw_total_sequences"]
            read_sum = (
                data["reads_mapped_MQ1"] + data["reads_MQ0"] + data["reads_unmapped"] + data["filtered_sequences"]
            )
            if read_sum == expected_total:
                bedgraph_data[sample_id] = data
            else:
                log.warning(
                    "sum of mapped/unmapped/filtered reads not matching total, "
                    "skipping samtools plot for: {}".format(sample_id)
                )
        self.add_section(
            name="Percent mapped",
            anchor="samtools-stats-alignment",
            description="Alignment metrics from <code>samtools stats</code>; mapped vs. unmapped reads vs. reads mapped with MQ0.",
            helptext="""
            For a set of samples that have come from the same multiplexed library,
            similar numbers of reads for each sample are expected. Large differences in numbers might
            indicate issues during the library preparation process. Whilst large differences in read
            numbers may be controlled for in downstream processings (e.g. read count normalisation),
            you may wish to consider whether the read depths achieved have fallen below recommended
            levels depending on the applications.

            Low alignment rates could indicate contamination of samples (e.g. adapter sequences),
            low sequencing quality or other artefacts. These can be further investigated in the
            sequence level QC (e.g. from FastQC).

            Reads mapped with MQ0 often indicate that the reads are ambiguously mapped to multiple
            locations in the reference sequence. This can be due to repetitive regions in the genome,
            the presence of alternative contigs in the reference, or due to reads that are too short
            to be uniquely mapped. These reads are often filtered out in downstream analyses.
            """,
            plot=alignment_chart(bedgraph_data),
        )


def alignment_chart(data):
    """Make the HighCharts HTML to plot the alignment rates"""
    keys = {
        "reads_mapped_MQ1": {"color": "#437bb1", "name": "Mapped (with MQ>0)"},
        "reads_MQ0": {"color": "#FF9933", "name": "MQ0"},
        "reads_unmapped": {"color": "#b1084c", "name": "Unmapped"},
    }

    # Config for the plot
    plot_conf = {
        "id": "samtools_alignment_plot",
        "title": "Samtools stats: Alignment Scores",
        "ylab": "# Reads",
        "cpswitch_counts_label": "Number of Reads",
    }
    return bargraph.plot(data, keys, plot_conf)



"""
=============
 sections
=============

This theme is very similar to 'default', except that
the output from each module is hidden by default.
Clicking the module names in the side-navigation
(on the left) hide the visible module and show the
module that was clicked on.

This approach is useful when reports are very large
and take a long time for the browser to render. By only
displaying one piece of the report at a time, the report
is easier to use.

"""
import os

from multiqc.utils import config

template_parent = "default"
template_dir = os.path.dirname(__file__)
base_fn = "base.html"

config.collapse_tables = False



#!/usr/bin/env python

""" MultiQC functions to plot a linegraph """

import base64
import io
import logging
import os
import random
import sys
from typing import Dict

from multiqc.utils import config, report, util_functions

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(plotdata, pconfig=None):
    if config.plots_force_flat or (
        not config.plots_force_interactive and plotdata and len(plotdata[0]) > config.plots_flat_numseries
    ):
        try:
            report.num_mpl_plots += 1
            return matplotlib_linegraph(plotdata, pconfig)
        except Exception as e:
            if config.strict:
                raise
            logger.error("############### Error making MatPlotLib figure! Falling back to HighCharts.")
            logger.debug(e, exc_info=True)
            return highcharts_linegraph(plotdata, pconfig)
    else:
        # Use MatPlotLib to generate static plots if requested
        if config.export_plots:
            matplotlib_linegraph(plotdata, pconfig)
        # Return HTML for HighCharts dynamic plot
        return highcharts_linegraph(plotdata, pconfig)


def highcharts_linegraph(plotdata, pconfig=None):
    """
    Build the HTML needed for a HighCharts line graph. Should be
    called by linegraph.plot(), which properly formats input data.
    """
    if pconfig is None:
        pconfig = {}

    # Get the plot ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_hcplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Build the HTML for the page
    html = '<div class="mqc_hcplot_plotgroup">'

    # Log Switch
    if pconfig.get("logswitch") is True:
        c_active = "active"
        l_active = ""
        if pconfig.get("logswitch_active") is True:
            c_active = ""
            l_active = "active"
        c_label = pconfig.get("cpswitch_counts_label", "Counts")
        l_label = pconfig.get("logswitch_label", "Log10")
        html += '<div class="btn-group hc_switch_group"> \n'
        html += '<button class="btn btn-default btn-sm {c_a}" data-action="set_numbers" data-target="{id}" data-ylab="{c_l}">{c_l}</button> \n'.format(
            id=pconfig["id"], c_a=c_active, c_l=c_label
        )
        if pconfig.get("logswitch") is True:
            html += '<button class="btn btn-default btn-sm {l_a}" data-action="set_log" data-target="{id}" data-ylab="{l_l}">{l_l}</button> \n'.format(
                id=pconfig["id"], l_a=l_active, l_l=l_label
            )
        html += "</div> "
        if len(plotdata) > 1:
            html += " &nbsp; &nbsp; "

    # Buttons to cycle through different datasets
    if len(plotdata) > 1:
        html += '<div class="btn-group hc_switch_group">\n'
        for k, p in enumerate(plotdata):
            active = "active" if k == 0 else ""
            try:
                name = pconfig["data_labels"][k]["name"]
            except Exception:
                try:
                    name = pconfig["data_labels"][k]
                except Exception:
                    name = k + 1
            try:
                ylab = f"data-ylab=\"{pconfig['data_labels'][k]['ylab']}\""
            except Exception:
                ylab = f'data-ylab="{name}"' if name != k + 1 else ""
            try:
                ymax = f"data-ymax=\"{pconfig['data_labels'][k]['ymax']}\""
            except Exception:
                ymax = ""
            try:
                xlab = f"data-xlab=\"{pconfig['data_labels'][k]['xlab']}\""
            except Exception:
                xlab = ""
            html += '<button class="btn btn-default btn-sm {a}" data-action="set_data" {y} {ym} {x} data-newdata="{k}" data-target="{id}">{n}</button>\n'.format(
                a=active, id=pconfig["id"], n=name, y=ylab, ym=ymax, x=xlab, k=k
            )
        html += "</div>\n\n"

    # The plot div
    html += '<div class="hc-plot-wrapper"{height}><div id="{id}" class="hc-plot not_rendered hc-line-plot"><small>loading..</small></div></div></div> \n'.format(
        id=pconfig["id"],
        height=f' style="height:{pconfig["height"]}px"' if "height" in pconfig else "",
    )

    report.num_hc_plots += 1

    report.plot_data[pconfig["id"]] = {"plot_type": "xy_line", "datasets": plotdata, "config": pconfig}

    return html


def matplotlib_linegraph(plotdata, pconfig=None):
    """
    Plot a line graph with Matplot lib and return a HTML string. Either embeds a base64
    encoded image within HTML or writes the plot and links to it. Should be called by
    plot_bargraph, which properly formats the input data.
    """
    try:
        # Import matplot lib but avoid default X environment
        import matplotlib

        matplotlib.use("Agg")
        import matplotlib.pyplot as plt

        logger.debug(f"Using matplotlib version {matplotlib.__version__}")
    except Exception as e:
        # MatPlotLib can break in a variety of ways. Fake an error message and continue without it if so.
        # The lack of the library will be handled when plots are attempted
        print("##### ERROR! MatPlotLib library could not be loaded!    #####", file=sys.stderr)
        print("##### Flat plots will instead be plotted as interactive #####", file=sys.stderr)
        print(e)

    if pconfig is None:
        pconfig = {}

    # Plot group ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_mplplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Individual plot IDs
    pids = []
    for k in range(len(plotdata)):
        try:
            name = pconfig["data_labels"][k]["name"]
        except Exception:
            name = k + 1
        pid = f"mqc_{pconfig['id']}_{name}"
        pid = report.save_htmlid(pid, skiplint=True)
        pids.append(pid)

    html = (
        '<p class="text-info"><small><span class="glyphicon glyphicon-picture" aria-hidden="true"></span> '
        + "Flat image plot. Toolbox functions such as highlighting / hiding samples will not work "
        + '(see the <a href="http://multiqc.info/docs/#flat--interactive-plots" target="_blank">docs</a>).</small></p>'
    )
    html += f"<div class=\"mqc_mplplot_plotgroup\" id=\"{pconfig['id']}\">"

    # Buttons to cycle through different datasets
    if len(plotdata) > 1 and not config.simple_output:
        html += '<div class="btn-group mpl_switch_group mqc_mplplot_bargraph_switchds">\n'
        for k, p in enumerate(plotdata):
            pid = pids[k]
            active = "active" if k == 0 else ""
            try:
                name = pconfig["data_labels"][k]["name"]
            except Exception:
                name = k + 1
            html += '<button class="btn btn-default btn-sm {a}" data-target="#{pid}">{n}</button>\n'.format(
                a=active, pid=pid, n=name
            )
        html += "</div>\n\n"

    # Go through datasets creating plots
    for pidx, pdata in enumerate(plotdata):
        # Plot ID
        pid = pids[pidx]

        # Save plot data to file
        if pconfig.get("save_data_file", True):
            fdata = dict()
            lastcats = None
            sharedcats = True
            for d in pdata:
                fdata[d["name"]] = dict()

                # Check to see if all categories are the same
                if len(d["data"]) > 0 and isinstance(d["data"][0], list):
                    if lastcats is None:
                        lastcats = [x[0] for x in d["data"]]
                    elif lastcats != [x[0] for x in d["data"]]:
                        sharedcats = False

                for i, x in enumerate(d["data"]):
                    if isinstance(x, list):
                        fdata[d["name"]][x[0]] = x[1]
                    else:
                        try:
                            fdata[d["name"]][pconfig["categories"][i]] = x
                        except Exception:
                            fdata[d["name"]][str(i)] = x

            # Custom tsv output if the x-axis varies
            if not sharedcats and config.data_format == "tsv":
                fout = ""
                for d in pdata:
                    fout += "\t" + "\t".join([str(x[0]) for x in d["data"]])
                    fout += f"\n{d['name']}\t"
                    fout += "\t".join([str(x[1]) for x in d["data"]])
                    fout += "\n"
                with io.open(os.path.join(config.data_dir, f"{pid}.txt"), "w", encoding="utf-8") as f:
                    print(fout.encode("utf-8", "ignore").decode("utf-8"), file=f)
            else:
                util_functions.write_data_file(fdata, pid)

        plt_height = 6
        # Use fixed height if pconfig['height'] is set (convert pixels -> inches)
        if "height" in pconfig:
            # Default interactive height in pixels = 512
            # Not perfect replication, but good enough
            plt_height = 6 * (pconfig["height"] / 512)

        # Set up figure
        fig = plt.figure(figsize=(14, plt_height), frameon=False)
        axes = fig.add_subplot(111)

        # Go through data series
        for idx, d in enumerate(pdata):
            # Line style
            linestyle = "solid"
            if d.get("dashStyle", None) == "Dash":
                linestyle = "dashed"

            # Reformat data (again)
            try:
                axes.plot(
                    [x[0] for x in d["data"]],
                    [x[1] for x in d["data"]],
                    label=d["name"],
                    color=d["color"],
                    linestyle=linestyle,
                    linewidth=1,
                    marker=None,
                )
            except TypeError:
                # Categorical data on x axis
                axes.plot(d["data"], label=d["name"], color=d["color"], linewidth=1, marker=None)

        # Log scale
        if pconfig.get("xLog", False):
            axes.set_xscale("log")
        if pconfig.get("yLog", False):
            axes.set_yscale("log")

        # Tidy up axes
        axes.tick_params(
            labelsize=pconfig.get("labelSize", 8), direction="out", left=False, right=False, top=False, bottom=False
        )
        axes.set_xlabel(pconfig.get("xlab", ""))
        axes.set_ylabel(pconfig.get("ylab", ""))

        # Dataset specific y label
        try:
            axes.set_ylabel(pconfig["data_labels"][pidx]["ylab"])
        except Exception:
            pass

        # Axis limits
        default_ylimits = axes.get_ylim()
        ymin = default_ylimits[0]
        if "ymin" in pconfig:
            ymin = pconfig["ymin"]
        elif "yFloor" in pconfig:
            ymin = max(pconfig["yFloor"], default_ylimits[0])
        ymax = default_ylimits[1]
        if "ymax" in pconfig:
            ymax = pconfig["ymax"]
        elif "yCeiling" in pconfig:
            ymax = min(pconfig["yCeiling"], default_ylimits[1])
        if (ymax - ymin) < pconfig.get("yMinRange", 0):
            ymax = ymin + pconfig["yMinRange"]
        axes.set_ylim((ymin, ymax))

        # Dataset specific ymax
        try:
            axes.set_ylim((ymin, pconfig["data_labels"][pidx]["ymax"]))
        except Exception:
            pass

        default_xlimits = axes.get_xlim()
        xmin = default_xlimits[0]
        if "xmin" in pconfig:
            xmin = pconfig["xmin"]
        elif "xFloor" in pconfig:
            xmin = max(pconfig["xFloor"], default_xlimits[0])
        xmax = default_xlimits[1]
        if "xmax" in pconfig:
            xmax = pconfig["xmax"]
        elif "xCeiling" in pconfig:
            xmax = min(pconfig["xCeiling"], default_xlimits[1])
        if (xmax - xmin) < pconfig.get("xMinRange", 0):
            xmax = xmin + pconfig.get("xMinRange", 0)
        axes.set_xlim((xmin, xmax))

        # Plot title
        if "title" in pconfig:
            plt.text(0.5, 1.05, pconfig["title"], horizontalalignment="center", fontsize=16, transform=axes.transAxes)
        axes.grid(True, zorder=10, which="both", axis="y", linestyle="-", color="#dedede", linewidth=1)

        # X axis categories, if specified
        if "categories" in pconfig:
            axes.set_xticks([i for i, v in enumerate(pconfig["categories"])])
            axes.set_xticklabels(pconfig["categories"])

        # Axis lines
        xlim = axes.get_xlim()
        axes.plot([xlim[0], xlim[1]], [0, 0], linestyle="-", color="#dedede", linewidth=2)
        axes.set_axisbelow(True)
        axes.spines["right"].set_visible(False)
        axes.spines["top"].set_visible(False)
        axes.spines["bottom"].set_visible(False)
        axes.spines["left"].set_visible(False)

        # Background colours, if specified
        if "yPlotBands" in pconfig:
            xlim = axes.get_xlim()
            for pb in pconfig["yPlotBands"]:
                axes.barh(
                    pb["from"],
                    xlim[1],
                    height=pb["to"] - pb["from"],
                    left=xlim[0],
                    color=pb["color"],
                    linewidth=0,
                    zorder=0,
                    align="edge",
                )
        if "xPlotBands" in pconfig:
            ylim = axes.get_ylim()
            for pb in pconfig["xPlotBands"]:
                axes.bar(
                    pb["from"],
                    ylim[1],
                    width=pb["to"] - pb["from"],
                    bottom=ylim[0],
                    color=pb["color"],
                    linewidth=0,
                    zorder=0,
                    align="edge",
                )

        # Tight layout - makes sure that legend fits in and stuff
        if len(pdata) <= 15:
            axes.legend(
                loc="lower center",
                bbox_to_anchor=(0, -0.22, 1, 0.102),
                ncol=5,
                mode="expand",
                fontsize=pconfig.get("labelSize", 8),
                frameon=False,
            )
            plt.tight_layout(rect=[0, 0.08, 1, 0.92])
        else:
            plt.tight_layout(rect=[0, 0, 1, 0.92])

        # Should this plot be hidden on report load?
        hidediv = ""
        if pidx > 0:
            hidediv = ' style="display:none;"'

        # Save the plot to the data directory if export is requests
        if config.export_plots:
            for fformat in config.export_plot_formats:
                # Make the directory if it doesn't already exist
                plot_dir = os.path.join(config.plots_dir, fformat)
                if not os.path.exists(plot_dir):
                    os.makedirs(plot_dir)
                # Save the plot
                plot_fn = os.path.join(plot_dir, f"{pid}.{fformat}")
                fig.savefig(plot_fn, format=fformat, bbox_inches="tight")

        # Output the figure to a base64 encoded string
        if getattr(get_template_mod(), "base64_plots", True) is True:
            img_buffer = io.BytesIO()
            fig.savefig(img_buffer, format="png", bbox_inches="tight")
            b64_img = base64.b64encode(img_buffer.getvalue()).decode("utf8")
            img_buffer.close()
            html += '<div class="mqc_mplplot" id="{}"{}><img src="data:image/png;base64,{}" /></div>'.format(
                pid, hidediv, b64_img
            )

        # Save to a file and link <img>
        else:
            plot_relpath = os.path.join(config.plots_dir_name, "png", f"{pid}.png")
            html += f'<div class="mqc_mplplot" id="{pid}"{hidediv}><img src="{plot_relpath}" /></div>'

        plt.close(fig)

    # Close wrapping div
    html += "</div>"

    return html


def smooth_line_data(data: Dict[str, Dict], numpoints: int) -> Dict[str, Dict[int, int]]:
    """
    Function to take an x-y dataset and use binning to smooth to a maximum number of datapoints.
    Each datapoint in a smoothed dataset corresponds to the first point in a bin.

    Examples to show the idea:

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=6
    we want to keep the first and the last element, thus excluding the last element from the binning:
    binsize = len([0 1 2 3 4 5 6 7 8]))/(numpoints-1) = 9/5 = 1.8
    taking points in indices rounded from multiples of 1.8: [0, 1.8, 3.6, 5.4, 7.2, 9],
    ...which evaluates to first_element_in_bin_indices=[0, 2, 4, 5, 7, 9]
    picking up the elements: [0 _ 2 _ 4 5 _ 7 _ 9]

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=9
    binsize = 9/8 = 1.125
    indices: [0.0, 1.125, 2.25, 3.375, 4.5, 5.625, 6.75, 7.875, 9] -> [0, 1, 2, 3, 5, 6, 7, 8, 9]
    picking up the elements: [0 1 2 3 _ 5 6 7 8 9]

    d=[0 1 2 3 4 5 6 7 8 9], numpoints=3
    binsize = len(d)/numpoints = 9/2 = 4.5
    indices: [0.0, 4.5, 9] -> [0, 5, 9]
    picking up the elements: [0 _ _ _ _ 5 _ _ _ 9]
    """
    smoothed_data = dict()
    for s_name, d in data.items():
        # Check that we need to smooth this data
        if len(d) <= numpoints or len(d) == 0:
            smoothed_data[s_name] = d
            continue

        binsize = (len(d) - 1) / (numpoints - 1)
        first_element_indices = [round(binsize * i) for i in range(numpoints)]
        smoothed_d = {x: y for i, (x, y) in enumerate(d.items()) if i in first_element_indices}
        smoothed_data[s_name] = smoothed_d

    return smoothed_data



""" MultiQC module to parse output from Cell Ranger count """

import json
import logging
import re

from multiqc import config
from multiqc.modules.cellranger.utils import set_hidden_cols, update_dict, parse_bcknee_data, transform_data
from multiqc.plots import bargraph, linegraph, table


# Initialise the logger
log = logging.getLogger(__name__)


class CellRangerCountMixin:
    """Cell Ranger count report parser"""

    def parse_count_html(self):
        self.cellrangercount_data = dict()
        self.cellrangercount_antibody_data = dict()
        self.cellrangercount_general_data = dict()
        self.cellrangercount_warnings = dict()
        self.cellrangercount_plots_conf = {"bc": dict(), "genes": dict()}
        self.cellrangercount_plots_data = {"bc": dict(), "genes": dict()}
        self.count_general_data_headers = dict()
        self.count_data_headers = dict()
        self.antibody_data_headers = dict()
        self.count_warnings_headers = dict()

        for f in self.find_log_files("cellranger/count_html", filehandles=True):
            self.parse_count_report(f)

        self.cellrangercount_data = self.ignore_samples(self.cellrangercount_data)
        if self.cellrangercount_antibody_data:
            self.cellrangercount_antibody_data = self.ignore_samples(self.cellrangercount_antibody_data)
        self.cellrangercount_general_data = self.ignore_samples(self.cellrangercount_general_data)
        self.cellrangercount_warnings = self.ignore_samples(self.cellrangercount_warnings)
        for k in self.cellrangercount_plots_data.keys():
            self.cellrangercount_plots_data[k] = self.ignore_samples(self.cellrangercount_plots_data[k])

        self.count_general_data_headers["reads"] = {
            "rid": "count_genstats_reads",
            "title": f"{config.read_count_prefix} Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "modify": lambda x: x * config.read_count_multiplier,
            "shared_key": "read_count",
            "namespace": "Count",
        }
        self.count_general_data_headers = set_hidden_cols(
            self.count_general_data_headers, ["Q30 bc", "Q30 UMI", "Q30 read"]
        )

        self.count_data_headers["reads"] = {
            "rid": "count_data_reads",
            "title": f"{config.read_count_prefix} Reads",
            "description": f"Number of reads ({config.read_count_desc})",
            "modify": lambda x: x * config.read_count_multiplier,
        }
        self.count_data_headers = set_hidden_cols(
            self.count_data_headers,
            [
                "Q30 bc",
                "Q30 UMI",
                "Q30 read",
                "reads in cells",
                "avg reads/cell",
                "confident reads",
                "confident transcriptome",
                "confident intronic",
                "confident intergenic",
                "reads antisense",
                "saturation",
            ],
        )

        if self.cellrangercount_antibody_data:
            self.antibody_data_headers["reads"] = {
                "rid": "antibody_data_reads",
                "title": f"{config.read_count_prefix} Reads",
                "description": f"Number of reads ({config.read_count_desc})",
                "modify": lambda x: x * config.read_count_multiplier,
            }
            self.antibody_data_headers = set_hidden_cols(
                self.antibody_data_headers,
                ["Q30 bc", "Q30 UMI", "Q30 read", "saturation", "umi per cell", "reads in aggregate bc"],
            )

        if len(self.cellrangercount_general_data) == 0:
            return 0

        self.general_stats_addcols(self.cellrangercount_general_data, self.count_general_data_headers)

        # Write parsed report data to a file
        self.write_data_file(self.cellrangercount_data, "multiqc_cellranger_count")
        if self.cellrangercount_antibody_data:
            self.write_data_file(self.cellrangercount_antibody_data, "multiqc_cellranger_antibody_count")

        # Add sections to the report
        if len(self.cellrangercount_warnings) > 0:
            self.add_section(
                name="Count - Warnings",
                anchor="cellranger-count-warnings",
                description="Warnings encountered during the analysis",
                plot=table.plot(
                    self.cellrangercount_warnings,
                    self.count_warnings_headers,
                    {
                        "namespace": "Count",
                        "id": "cellranger-count-warnings-table",
                        "title": "Cellranger Count: Warnings",
                    },
                ),
            )

        self.add_section(
            name="Count - Summary stats",
            anchor="cellranger-count-stats",
            description="Summary QC metrics from Cell Ranger count",
            plot=table.plot(
                self.cellrangercount_data,
                self.count_data_headers,
                {
                    "namespace": "Count",
                    "id": "cellranger-count-stats-table",
                    "title": "Cellranger Count: Summary Stats",
                },
            ),
        )

        self.add_section(
            name="Count - BC rank plot",
            anchor="cellranger-count-bcrank-plot",
            description=self.cellrangercount_plots_conf["bc"]["description"],
            helptext=self.cellrangercount_plots_conf["bc"]["helptext"],
            plot=linegraph.plot(self.cellrangercount_plots_data["bc"], self.cellrangercount_plots_conf["bc"]["config"]),
        )

        self.add_section(
            name="Count - Median genes",
            anchor="cellranger-count-genes-plot",
            description=self.cellrangercount_plots_conf["genes"]["description"],
            helptext=self.cellrangercount_plots_conf["genes"]["helptext"],
            plot=linegraph.plot(
                self.cellrangercount_plots_data["genes"], self.cellrangercount_plots_conf["genes"]["config"]
            ),
        )

        if "saturation" in self.cellrangercount_plots_data:
            self.add_section(
                name="Count - Saturation plot",
                anchor="cellranger-count-saturation-plot",
                description=self.cellrangercount_plots_conf["saturation"]["description"],
                helptext=self.cellrangercount_plots_conf["saturation"]["helptext"],
                plot=linegraph.plot(
                    self.cellrangercount_plots_data["saturation"],
                    self.cellrangercount_plots_conf["saturation"]["config"],
                ),
            )

        if self.cellrangercount_antibody_data:
            self.add_section(
                name="Antibody - Summary stats",
                anchor="cellranger-antibody-stats",
                description="Summary QC metrics from Cell Ranger count",
                plot=table.plot(
                    self.cellrangercount_antibody_data,
                    self.antibody_data_headers,
                    {
                        "namespace": "Antibody",
                        "id": "cellranger-antibody-stats-table",
                        "title": "Cellranger Antibody: Summary Stats",
                    },
                ),
            )

        if "antibody_counts" in self.cellrangercount_plots_conf:
            self.add_section(
                name="Antibody - Counts Distribution Bargraph",
                anchor="cellranger-antibody-counts",
                description=self.cellrangercount_plots_conf["antibody_counts"]["description"],
                helptext=self.cellrangercount_plots_conf["antibody_counts"]["helptext"],
                plot=bargraph.plot(
                    self.cellrangercount_plots_data["antibody_counts"],
                    self.cellrangercount_plots_conf["antibody_counts"]["keys"],
                    self.cellrangercount_plots_conf["antibody_counts"]["config"],
                ),
            )

        return len(self.cellrangercount_general_data)

    def parse_count_report(self, f):
        """Go through the html report of cell ranger and extract the data in a dicts"""

        for line in f["f"]:
            line = line.strip()
            if line.startswith("const data"):
                line = line.replace("const data = ", "")
                summary = json.loads(line)
                summary = summary["summary"]
                break

        s_name = self.clean_s_name(summary["sample"]["id"], f)

        # Extract software version
        try:
            version_pair = summary["summary_tab"]["pipeline_info_table"]["rows"][-1]
            assert version_pair[0] == "Pipeline Version"
            version_match = re.search(r"cellranger-([\d\.]+)", version_pair[1])
            if version_match:
                self.add_software_version(version_match.group(1), s_name)
        except (KeyError, AssertionError):
            log.debug(f"Unable to parse version for sample {s_name}")

        data_general_stats = dict()

        # Store general stats from cells
        col_dict = {
            "Estimated Number of Cells": "estimated cells",
            "Mean Reads per Cell": "avg reads/cell",
            "Fraction Reads in Cells": "reads in cells",
        }
        colours = {
            "estimated cells": "PuBu",
            "avg reads/cell": "GnBu",
            "reads in cells": "Purples",
        }
        data_general_stats, self.count_general_data_headers = update_dict(
            data_general_stats,
            self.count_general_data_headers,
            summary["summary_tab"]["cells"]["table"]["rows"],
            col_dict,
            colours,
            "Count",
        )

        # Store general stats from sequencing tables
        col_dict = {
            "Number of Reads": "reads",
            "Valid Barcodes": "valid bc",
            "Q30 Bases in Barcode": "Q30 bc",
            "Q30 Bases in UMI": "Q30 UMI",
            "Q30 Bases in RNA Read": "Q30 read",
        }
        colours = {
            "reads": "PuBuGn",
            "valid bc": "RdYlGn",
            "Q30 bc": "RdYlBu",
            "Q30 UMI": "Spectral",
            "Q30 read": "RdBu",
        }
        data_general_stats, self.count_general_data_headers = update_dict(
            data_general_stats,
            self.count_general_data_headers,
            summary["summary_tab"]["sequencing"]["table"]["rows"],
            col_dict,
            colours,
            "Count",
        )

        # Store full data from cell ranger count report
        data_rows = (
            summary["summary_tab"]["sequencing"]["table"]["rows"]
            + summary["summary_tab"]["cells"]["table"]["rows"]
            + summary["summary_tab"]["mapping"]["table"]["rows"]
        )
        col_dict = {
            "Number of Reads": "reads",
            "Estimated Number of Cells": "estimated cells",
            "Mean Reads per Cell": "avg reads/cell",
            "Total Genes Detected": "genes detected",
            "Median Genes per Cell": "median genes/cell",
            "Fraction Reads in Cells": "reads in cells",
            "Valid Barcodes": "valid bc",
            "Valid UMIs": "valid umi",
            "Median UMI Counts per Cell": "median umi/cell",
            "Sequencing Saturation": "saturation",
            "Q30 Bases in Barcode": "Q30 bc",
            "Q30 Bases in UMI": "Q30 UMI",
            "Q30 Bases in RNA Read": "Q30 read",
            "Reads Mapped to Genome": "reads mapped",
            "Reads Mapped Confidently to Genome": "confident reads",
            "Reads Mapped Confidently to Transcriptome": "confident transcriptome",
            "Reads Mapped Confidently to Exonic Regions": "confident exonic",
            "Reads Mapped Confidently to Intronic Regions": "confident intronic",
            "Reads Mapped Confidently to Intergenic Regions": "confident intergenic",
            "Reads Mapped Antisense to Gene": "reads antisense",
        }
        colours = {
            "reads": "YlGn",
            "estimated cells": "RdPu",
            "avg reads/cell": "Blues",
            "genes detected": "Greens",
            "median genes/cell": "Purples",
            "reads in cells": "PuBuGn",
            "valid bc": "Spectral",
            "valid umi": "RdYlGn",
            "median umi/cell": "YlGn",
            "saturation": "YlOrRd",
        }
        table, self.count_data_headers = update_dict(
            data_general_stats,
            self.count_data_headers,
            data_rows,
            col_dict,
            colours,
            "Count",
        )
        if not table:
            return None

        # Extract warnings if any
        warnings = dict()
        alarms_list = summary["alarms"].get("alarms", [])
        for alarm in alarms_list:
            # "Intron mode used" alarm added in Cell Ranger 7.0 lacks id
            if "id" not in alarm:
                continue
            warnings[alarm["id"]] = "FAIL"
            self.count_warnings_headers[alarm["id"]] = {
                "title": alarm["id"],
                "description": alarm["title"],
                "bgcols": {"FAIL": "#f06807"},
            }

        # Extract data for plots
        help_dict = {x[0]: x[1][0] for x in summary["summary_tab"]["cells"]["help"]["data"]}
        plots = {
            "bc": {
                "config": {
                    "id": "mqc_cellranger_count_bc_knee",
                    "title": f"Cell Ranger count: {summary['summary_tab']['cells']['barcode_knee_plot']['layout']['title']}",
                    "xlab": summary["summary_tab"]["cells"]["barcode_knee_plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["summary_tab"]["cells"]["barcode_knee_plot"]["layout"]["yaxis"]["title"],
                    "yLog": True,
                    "xLog": True,
                },
                "description": "Barcode knee plot",
                "helptext": help_dict["Barcode Rank Plot"],
            },
            "genes": {
                "config": {
                    "id": "mqc_cellranger_count_genesXcell",
                    "title": f"Cell Ranger count: {summary['analysis_tab']['median_gene_plot']['help']['title']}",
                    "xlab": summary["analysis_tab"]["median_gene_plot"]["plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["analysis_tab"]["median_gene_plot"]["plot"]["layout"]["yaxis"]["title"],
                    "yLog": False,
                    "xLog": False,
                },
                "description": "Median gene counts per cell",
                "helptext": summary["analysis_tab"]["median_gene_plot"]["help"]["helpText"],
            },
        }
        try:
            plots["saturation"] = {
                "config": {
                    "id": "mqc_cellranger_count_saturation",
                    "title": f"Cell Ranger count: {summary['analysis_tab']['seq_saturation_plot']['help']['title']}",
                    "xlab": summary["analysis_tab"]["seq_saturation_plot"]["plot"]["layout"]["xaxis"]["title"],
                    "ylab": summary["analysis_tab"]["seq_saturation_plot"]["plot"]["layout"]["yaxis"]["title"],
                    "yLog": False,
                    "xLog": False,
                    "ymin": 0,
                    "ymax": 1,
                },
                "description": "Sequencing saturation",
                "helptext": summary["analysis_tab"]["seq_saturation_plot"]["help"]["helpText"],
            }
        except KeyError:
            pass

        plots_data = {
            "bc": parse_bcknee_data(summary["summary_tab"]["cells"]["barcode_knee_plot"]["data"], s_name),
            "genes": {s_name: transform_data(summary["analysis_tab"]["median_gene_plot"]["plot"]["data"][0])},
        }
        if "seq_saturation_plot" in summary["analysis_tab"]:
            plots_data["saturation"] = {
                s_name: transform_data(summary["analysis_tab"]["seq_saturation_plot"]["plot"]["data"][0])
            }

        # Store full data for ANTIBODY capture
        antibody_data = dict()
        if "ANTIBODY_sequencing" in summary["summary_tab"]:
            data_rows = (
                summary["summary_tab"]["ANTIBODY_sequencing"]["table"]["rows"]
                + summary["summary_tab"]["ANTIBODY_application"]["table"]["rows"]
            )
            col_dict = {
                "Number of Reads": "reads",
                "Valid Barcodes": "valid bc",
                "Valid UMIs": "valid umi",
                "Sequencing Saturation": "saturation",
                "Q30 Bases in Barcode": "Q30 bc",
                "Q30 Bases in Antibody Read": "Q30 read",
                "Q30 Bases in UMI": "Q30 UMI",
                "Fraction Antibody Reads": "antibody reads",
                "Fraction Antibody Reads Usable": "antibody reads usable",
                "Antibody Reads Usable per Cell": "antibody reads usable/cell",
                "Fraction Antibody Reads in Aggregate Barcodes": "reads in aggregate bc",
                "Fraction Unrecognized Antibody": "unrecognized antibody",
                "Antibody Reads in Cells": "antibody reads in cells",
                "Median UMIs per Cell (summed over all recognized antibody barcodes)": "umi per cell",
            }
            colours = {
                "reads": "YlGn",
                "antibody reads": "RdPu",
                "reads in cells": "Blues",
                "reads usable": "Greens",
                "reads usable per cell": "Purples",
                "reads in aggregate bc": "PuBuGn",
                "valid bc": "Spectral",
                "valid umi": "RdYlGn",
                "Q30 bc": "YlGn",
                "saturation": "YlOrRd",
            }
            antibody_data, self.antibody_data_headers = update_dict(
                antibody_data,
                self.antibody_data_headers,
                data_rows,
                col_dict,
                colours,
                "Antibody",
            )

            # Extract labels and values for the bargraph data
            combined_data = {}
            for label, value in zip(
                summary["antibody_tab"]["antibody_treemap_plot"]["plot"]["data"][0]["labels"],
                summary["antibody_tab"]["antibody_treemap_plot"]["plot"]["data"][0]["values"],
            ):
                label_match = re.search(r"<b>(.*?)\s+\((.*?)%\)</b>", label)
                if label_match:
                    label_value = label_match.group(1)
                    value_ = round(value * 100, 2)
                    combined_data[label_value] = value_

            # Extract labels and number of cells for labelling the bargraph
            combined_label = {}
            for label, cells in zip(
                summary["antibody_tab"]["antibody_treemap_plot"]["plot"]["data"][0]["labels"],
                summary["antibody_tab"]["antibody_treemap_plot"]["plot"]["data"][0]["text"],
            ):
                label_match = re.search(r"<b>(.*?)\s+\((.*?)%\)</b>", label)
                if label_match:
                    label_value = label_match.group(1)
                    combined_label[label_value] = label_value + ": " + cells

            # Use the label from `combined_label` for the plot
            keys = dict()
            for key, value in combined_label.items():
                keys[key] = {"name": value}

            plots["antibody_counts"] = {
                "config": {
                    "id": "mqc_cellranger_antibody_counts",
                    "title": "Cell Ranger: Distribution of Antibody Counts",
                    "ylab": "% Total UMI",
                    "ymax": 100,
                    "cpswitch": False,
                    "use_legend": False,
                    "tt_decimals": 2,
                    "tt_suffix": "%",
                    "tt_percentages": False,
                },
                "keys": keys,
                "description": "Antibody Counts Distribution Plot",
                "helptext": "Relative composition of antibody counts for features with at least 1 UMI. Box size represents fraction of total UMIs from cell barcodes that are derived from this antibody. Hover over a box to view more information on a particular antibody, including number of associated barcodes.",
            }
            plots_data["antibody_counts"] = {s_name: combined_data}

        if s_name in self.cellrangercount_general_data:
            log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
        self.add_data_source(f, s_name, module="cellranger", section="count")
        self.cellrangercount_data[s_name] = table
        if "antibody_tab" in summary:
            self.cellrangercount_antibody_data[s_name] = antibody_data
        self.cellrangercount_general_data[s_name] = data_general_stats
        if len(warnings) > 0:
            self.cellrangercount_warnings[s_name] = warnings
        self.cellrangercount_plots_conf.update(plots)
        for k in plots_data.keys():
            if k not in self.cellrangercount_plots_data.keys():
                self.cellrangercount_plots_data[k] = dict()
            self.cellrangercount_plots_data[k].update(plots_data[k])



""" MultiQC submodule to parse output from Bcftools stats """

import logging
import re

from multiqc import config
from multiqc.plots import bargraph, linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"# This file was produced by bcftools stats \(([\d\.]+)"
HTSLIB_REGEX = r"\+htslib-([\d\.]+)"


class StatsReportMixin:
    """Mixin loaded by the bcftools MultiqcModule class"""

    def parse_bcftools_stats(self):
        """
        Find bcftools stats logs and parse their data
          Bcftools stats reports contain 'sets' of data, which can
          have multiple vcf files each (but usually don't). Here,
          we treat each 'set' as a MultiQC sample, taking the first
          input filename for each set as the name.
        """
        collapse_complementary = getattr(config, "bcftools", {}).get("collapse_complementary_changes", False)
        if collapse_complementary:
            types = ["A>C", "A>G", "A>T", "C>A", "C>G", "C>T"]
        else:
            types = ["A>C", "A>G", "A>T", "C>A", "C>G", "C>T", "G>A", "G>C", "G>T", "T>A", "T>C", "T>G"]

        self.bcftools_stats = dict()
        self.bcftools_stats_indels = dict()
        self.bcftools_stats_vqc_snp = dict()
        self.bcftools_stats_sample_variants = dict()
        self.bcftools_stats_sample_tstv = dict()
        self.bcftools_stats_sample_singletons = dict()
        self.bcftools_stats_sample_depth = dict()
        self.bcftools_stats_vqc_transi = dict()
        self.bcftools_stats_vqc_transv = dict()
        self.bcftools_stats_vqc_indels = dict()
        self.bcftools_stats_depth_data = dict()
        for f in self.find_log_files("bcftools/stats"):
            s_names = list()
            for line in f["f"].splitlines():
                # Get version number from file contents
                if line.startswith("# This file was produced by bcftools stats"):
                    # Look for BCFtools version
                    version_match = re.search(VERSION_REGEX, line)
                    if version_match is None:
                        continue

                    # Add BCFtools version
                    bcftools_version = version_match.group(1)
                    self.add_software_version(bcftools_version, f["s_name"])

                    # Look for HTSlib version
                    htslib_version_match = re.search(HTSLIB_REGEX, line)
                    if htslib_version_match is None:
                        continue

                    # Add HTSlib version if different from BCFtools version
                    htslib_version = htslib_version_match.group(1)
                    if htslib_version != bcftools_version:
                        self.add_software_version(htslib_version, f["s_name"], "HTSlib")

                s = line.split("\t")
                # Get the sample names - one per 'set'
                if s[0] == "ID":
                    s_name = self.clean_s_name(s[2], f)
                    s_names.append(s_name)
                    if s_name in self.bcftools_stats:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name, section="stats")
                    self.bcftools_stats[s_name] = dict()
                    self.bcftools_stats_indels[s_name] = dict()
                    self.bcftools_stats_sample_variants[s_name] = dict()
                    self.bcftools_stats_sample_tstv[s_name] = dict()
                    self.bcftools_stats_sample_singletons[s_name] = dict()
                    self.bcftools_stats_sample_depth[s_name] = dict()
                    self.bcftools_stats_vqc_snp[s_name] = dict()
                    self.bcftools_stats_vqc_transi[s_name] = dict()
                    self.bcftools_stats_vqc_transv[s_name] = dict()
                    self.bcftools_stats_vqc_indels[s_name] = dict()
                    self.bcftools_stats_depth_data[s_name] = {}
                    self.bcftools_stats_indels[s_name][0] = None  # Avoid joining line across missing 0

                # Parse key stats
                if s[0] == "SN" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    field = s[2].strip()[:-1]
                    field = field.replace(" ", "_")
                    value = int(s[3].strip())
                    self.bcftools_stats[s_name][field] = value

                # Parse transitions/transversions stats
                if s[0] == "TSTV" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    fields = ["ts", "tv", "tstv", "ts_1st_ALT", "tv_1st_ALT", "tstv_1st_ALT"]
                    for i, field in enumerate(fields):
                        value = s[i + 2].strip()
                        if "tstv" in field:
                            value = float(value)
                        else:
                            value = int(value)
                        self.bcftools_stats[s_name][field] = value

                # Parse substitution types
                if s[0] == "ST" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]

                    rc = {"A": "T", "C": "G", "G": "C", "T": "A"}
                    change = s[2].strip()
                    if change not in types:
                        change = ">".join(rc[n] for n in change.split(">"))

                    field = f"substitution_type_{change}"
                    value = int(s[3].strip())
                    if field not in self.bcftools_stats[s_name]:
                        self.bcftools_stats[s_name][field] = 0
                    self.bcftools_stats[s_name][field] += value

                # Indel length distributions
                if s[0] == "IDD" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    length = int(s[2].strip())
                    count = int(s[3].strip())
                    self.bcftools_stats_indels[s_name][length] = count

                # Per-sample counts
                if s[0] == "PSC" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    fields = ["variations_hom", "variations_het"]
                    for i, field in enumerate(fields):
                        self.bcftools_stats[s_name][field] = int(s[i + 4].strip())

                # Per-sample variant stats
                if s[0] == "PSC" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    sample = self.clean_s_name(s[2].strip(), f)
                    self.bcftools_stats_sample_variants[s_name][sample] = dict()
                    self.bcftools_stats_sample_variants[s_name][sample]["nSNPs"] = (
                        int(s[3].strip()) + int(s[4].strip()) + int(s[5].strip())
                    )
                    self.bcftools_stats_sample_variants[s_name][sample]["nIndels"] = int(s[8].strip())
                    if len(s) >= 14:
                        nMissing = int(s[13].strip())
                    else:
                        nMissing = 0
                    nPresent = self.bcftools_stats[s_name]["number_of_records"] - nMissing
                    self.bcftools_stats_sample_variants[s_name][sample]["nOther"] = (
                        nPresent
                        - self.bcftools_stats_sample_variants[s_name][sample]["nSNPs"]
                        - self.bcftools_stats_sample_variants[s_name][sample]["nIndels"]
                    )

                # Per-sample ts/tv stats
                if s[0] == "PSC" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    sample = self.clean_s_name(s[2].strip(), f)
                    self.bcftools_stats_sample_tstv[s_name][sample] = dict()
                    if int(s[7].strip()) != 0:
                        self.bcftools_stats_sample_tstv[s_name][sample]["tstv"] = float(
                            int(s[6].strip()) / int(s[7].strip())
                        )
                    else:
                        self.bcftools_stats_sample_tstv[s_name][sample]["tstv"] = 0

                # Per-sample singletons stats
                if s[0] == "PSC" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    sample = self.clean_s_name(s[2].strip(), f)
                    self.bcftools_stats_sample_singletons[s_name][sample] = dict()
                    self.bcftools_stats_sample_singletons[s_name][sample]["singletons"] = int(s[10].strip())
                    self.bcftools_stats_sample_singletons[s_name][sample]["rest"] = (
                        int(s[3].strip()) + int(s[4].strip()) + int(s[5].strip()) - int(s[10].strip())
                    )

                # Per-sample coverage stats
                if s[0] == "PSC" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    sample = self.clean_s_name(s[2].strip(), f)
                    self.bcftools_stats_sample_depth[s_name][sample] = dict()
                    self.bcftools_stats_sample_depth[s_name][sample]["depth"] = float(s[9].strip())

                # Depth plots
                if s[0] == "DP" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    bin_name = s[2].strip()
                    percent_sites = float(s[-1].strip())
                    self.bcftools_stats_depth_data[s_name][bin_name] = percent_sites

                # Variant Qualities
                if s[0] == "QUAL" and len(s_names) > 0:
                    s_name = s_names[int(s[1])]
                    quality = float("0" + s[2].strip())
                    self.bcftools_stats_vqc_snp[s_name][quality] = float(s[3].strip())
                    self.bcftools_stats_vqc_transi[s_name][quality] = float(s[4].strip())
                    self.bcftools_stats_vqc_transv[s_name][quality] = float(s[5].strip())
                    self.bcftools_stats_vqc_indels[s_name][quality] = float(s[6].strip())

        # Remove empty samples
        self.bcftools_stats = {k: v for k, v in self.bcftools_stats.items() if len(v) > 0}
        self.bcftools_stats_indels = {k: v for k, v in self.bcftools_stats_indels.items() if len(v) > 0}
        self.bcftools_stats_sample_variants = {
            k: v for k, v in self.bcftools_stats_sample_variants.items() if len(v) > 0
        }
        self.bcftools_stats_sample_tstv = {k: v for k, v in self.bcftools_stats_sample_tstv.items() if len(v) > 0}
        self.bcftools_stats_sample_singletons = {
            k: v for k, v in self.bcftools_stats_sample_singletons.items() if len(v) > 0
        }
        self.bcftools_stats_sample_depth = {k: v for k, v in self.bcftools_stats_sample_depth.items() if len(v) > 0}
        self.bcftools_stats_vqc_snp = {k: v for k, v in self.bcftools_stats_vqc_snp.items() if len(v) > 0}
        self.bcftools_stats_vqc_transi = {k: v for k, v in self.bcftools_stats_vqc_transi.items() if len(v) > 0}
        self.bcftools_stats_vqc_transv = {k: v for k, v in self.bcftools_stats_vqc_transv.items() if len(v) > 0}
        self.bcftools_stats_vqc_indels = {k: v for k, v in self.bcftools_stats_vqc_indels.items() if len(v) > 0}
        self.bcftools_stats_depth_data = {k: v for k, v in self.bcftools_stats_depth_data.items() if len(v) > 0}

        # Filter to strip out ignored sample names
        self.bcftools_stats = self.ignore_samples(self.bcftools_stats)
        self.bcftools_stats_indels = self.ignore_samples(self.bcftools_stats_indels)
        self.bcftools_stats_vqc_snp = self.ignore_samples(self.bcftools_stats_vqc_snp)
        self.bcftools_stats_sample_variants = self.ignore_samples(self.bcftools_stats_sample_variants)
        self.bcftools_stats_sample_tstv = self.ignore_samples(self.bcftools_stats_sample_tstv)
        self.bcftools_stats_sample_singletons = self.ignore_samples(self.bcftools_stats_sample_singletons)
        self.bcftools_stats_sample_depth = self.ignore_samples(self.bcftools_stats_sample_depth)
        self.bcftools_stats_vqc_transi = self.ignore_samples(self.bcftools_stats_vqc_transi)
        self.bcftools_stats_vqc_transv = self.ignore_samples(self.bcftools_stats_vqc_transv)
        self.bcftools_stats_vqc_indels = self.ignore_samples(self.bcftools_stats_vqc_indels)
        self.bcftools_stats_depth_data = self.ignore_samples(self.bcftools_stats_depth_data)
        if len(self.bcftools_stats) == 0:
            return 0

        # Write parsed report data to a file
        self.write_data_file(self.bcftools_stats, "multiqc_bcftools_stats")

        # Stats Table
        stats_headers = self.bcftools_stats_genstats_headers()
        if getattr(config, "bcftools", {}).get("write_general_stats", True):
            self.general_stats_addcols(self.bcftools_stats, stats_headers, "Stats")
        if getattr(config, "bcftools", {}).get("write_separate_table", False):
            self.add_section(
                name="Bcftools Stats",
                anchor="bcftools-stats_stats",
                plot=table.plot(
                    self.bcftools_stats,
                    stats_headers,
                    {
                        "namespace": "Stats",
                        "id": "bcftools-stats-table",
                    },
                ),
            )

        # Make bargraph plot of substitution types
        keys = {}
        for t in types:
            keys[f"substitution_type_{t}"] = {"name": t}
        pconfig = {
            "id": "bcftools-stats-subtypes",
            "title": "Bcftools Stats: Substitutions",
            "ylab": "# Substitutions",
            "cpswitch_counts_label": "Number of Substitutions",
        }
        self.add_section(
            name="Variant Substitution Types",
            anchor="bcftools-stats_variant_sub_types",
            plot=bargraph.plot(self.bcftools_stats, keys, pconfig),
        )

        # Make histograms of variant quality
        if len(self.bcftools_stats_vqc_snp) > 0:
            pconfig = {
                "id": "bcftools_stats_vqc",
                "title": "Bcftools Stats: Variant Quality Count",
                "ylab": "Count",
                "xlab": "Quality",
                "xDecimals": False,
                "ymin": 0,
                "smooth_points": 600,
                "data_labels": [
                    "Count SNP",
                    "Count Transitions",
                    "Count Transversions",
                    "Count Indels",
                ],
            }
            self.add_section(
                name="Variant Quality",
                anchor="bcftools-stats_variant_quality_plot",
                plot=linegraph.plot(
                    [
                        self.bcftools_stats_vqc_snp,
                        self.bcftools_stats_vqc_transi,
                        self.bcftools_stats_vqc_transv,
                        self.bcftools_stats_vqc_indels,
                    ],
                    pconfig,
                ),
            )

        # Make line graph of indel lengths
        if len(self.bcftools_stats_indels) > 0:
            pconfig = {
                "id": "bcftools_stats_indel-lengths",
                "title": "Bcftools Stats: Indel Distribution",
                "ylab": "Count",
                "xlab": "InDel Length (bp)",
                "xDecimals": False,
                "ymin": 0,
            }
            self.add_section(
                name="Indel Distribution",
                anchor="bcftools-stats_indel_plot",
                plot=linegraph.plot(self.bcftools_stats_indels, pconfig),
            )
        # Make line graph of variants per depth
        if len(self.bcftools_stats_depth_data) > 0:
            # Get shared list of bins and order them numerically
            all_bins = []
            for sname, val_by_bin in self.bcftools_stats_depth_data.items():
                all_bins.extend(list(val_by_bin.keys()))
            all_bins = sorted(all_bins, key=lambda x: int(re.sub(r"\D", "", x)))
            # Order bins in samples and fill missing bins:
            sorted_data = {
                sname: {b: self.bcftools_stats_depth_data[sname].get(b, 0) for b in all_bins}
                for sname in self.bcftools_stats_depth_data
            }
            pconfig = {
                "id": "bcftools_stats_variant_depths",
                "title": "Bcftools Stats: Variant depths",
                "ylab": "Fraction of sites (%)",
                "xlab": "Variant depth",
                "ymin": 0,
                "ymax": 100,
                "categories": True,
                "tt_decimals": 1,
            }
            self.add_section(
                name="Variant depths",
                anchor="bcftools-stats_depth_plot",
                description="Read depth support distribution for called variants",
                plot=linegraph.plot(sorted_data, pconfig),
            )

        # Make bargraph plot of missing sites
        if len(self.bcftools_stats_sample_variants) > 0:
            pconfig = {
                "id": "bcftools-stats-sites",
                "title": "Bcftools Stats: Sites per sample",
                "ylab": "# Sites",
                "cpswitch_counts_label": "Number of sites",
                "data_labels": list(self.bcftools_stats_sample_variants),
            }
            self.add_section(
                name="Sites per sample",
                anchor="bcftools-stats_sites_per_sample",
                plot=bargraph.plot(
                    list(self.bcftools_stats_sample_variants.values()), ["nSNPs", "nIndels", "nOther"], pconfig
                ),
            )

        # Make bargraph plot of ts/tv stats
        if len(self.bcftools_stats_sample_tstv) > 0:
            pconfig = {
                "id": "bcftools-stats-tstv",
                "title": "Bcftools Stats: Ts/Tv",
                "ylab": "Ts/Tv",
                "cpswitch_counts_label": "Ts/Tv",
                "cpswitch": False,
                "data_labels": list(self.bcftools_stats_sample_tstv),
            }
            self.add_section(
                name="Ts/Tv",
                anchor="bcftools-stats_ts_tv",
                plot=bargraph.plot(list(self.bcftools_stats_sample_tstv.values()), ["tstv"], pconfig),
            )

        # Make bargraph plot of singletons stats
        if len(self.bcftools_stats_sample_singletons) > 0:
            pconfig = {
                "id": "bcftools-stats-singletons",
                "title": "Bcftools Stats: Singletons",
                "ylab": "# Singletons",
                "cpswitch_counts_label": "Singletons",
                "cpswitch_c_active": False,
                "data_labels": list(self.bcftools_stats_sample_singletons),
            }
            self.add_section(
                name="Number of Singletons",
                anchor="bcftools-stats_singletones",
                plot=bargraph.plot(
                    list(self.bcftools_stats_sample_singletons.values()), ["singletons", "rest"], pconfig
                ),
            )

        # Make bargraph plot of sequencing depth stats
        if len(self.bcftools_stats_sample_depth) > 0:
            pconfig = {
                "id": "bcftools-stats-sequencing-depth",
                "title": "Bcftools Stats: Sequencing depth",
                "ylab": "Sequencing depth",
                "cpswitch_counts_label": "Sequencing depth",
                "cpswitch": False,
                "data_labels": list(self.bcftools_stats_sample_depth),
            }
            self.add_section(
                name="Sequencing depth",
                anchor="bcftools-stats_sequencing_depth",
                plot=bargraph.plot(list(self.bcftools_stats_sample_depth.values()), ["depth"], pconfig),
            )

        # Return the number of logs that were found
        return len(self.bcftools_stats)

    @staticmethod
    def bcftools_stats_genstats_headers():
        """Add key statistics to the General Stats table"""
        stats_headers = {
            "number_of_records": {
                "title": "Vars",
                "description": "Variations total",
                "min": 0,
                "format": "{:,.0f}",
            },
            "variations_hom": {
                "title": "Hom",
                "description": "Variations homozygous",
                "min": 0,
                "format": "{:,.0f}",
            },
            "variations_het": {
                "title": "Het",
                "description": "Variations heterozygous",
                "min": 0,
                "format": "{:,.0f}",
            },
            "number_of_SNPs": {
                "title": "SNP",
                "description": "Variation SNPs",
                "min": 0,
                "format": "{:,.0f}",
            },
            "number_of_indels": {
                "title": "Indel",
                "description": "Variation Insertions/Deletions",
                "min": 0,
                "format": "{:,.0f}",
            },
            "tstv": {
                "title": "Ts/Tv",
                "description": "Variant SNP transition / transversion ratio",
                "min": 0,
                "format": "{:,.2f}",
            },
            "number_of_MNPs": {
                "title": "MNP",
                "description": "Variation multinucleotide polymorphisms",
                "min": 0,
                "format": "{:,.0f}",
                "hidden": True,
            },
        }
        return stats_headers



""" MultiQC module to parse output from Picard """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Import the Picard submodules
from . import (
    AlignmentSummaryMetrics,
    BaseDistributionByCycleMetrics,
    IlluminaBasecallingMetrics,
    IlluminaLaneMetrics,
    CrosscheckFingerprints,
    ExtractIlluminaBarcodes,
    GcBiasMetrics,
    HsMetrics,
    InsertSizeMetrics,
    MarkDuplicates,
    MarkIlluminaAdapters,
    OxoGMetrics,
    QualityByCycleMetrics,
    QualityScoreDistributionMetrics,
    QualityYieldMetrics,
    RnaSeqMetrics,
    RrbsSummaryMetrics,
    TargetedPcrMetrics,
    ValidateSamFile,
    VariantCallingMetrics,
    WgsMetrics,
)

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    Picard is a collection of scripts. This MultiQC module supports some but not all.
    The code for each script is split into its own file and adds a section to the
    module output if logs are found.
    """

    def __init__(
        self,
        name="Picard",
        anchor="picard",
        href="http://broadinstitute.github.io/picard/",
        info="is a set of Java command line tools for manipulating high-throughput sequencing data.",
        # No DOI to cite // doi=
    ):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name=name,
            anchor=anchor,
            href=href,
            info=info,
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        for tool, mod in {
            "AlignmentSummaryMetrics": AlignmentSummaryMetrics,
            "BaseDistributionByCycleMetrics": BaseDistributionByCycleMetrics,
            "CrosscheckFingerprints": CrosscheckFingerprints,
            "GcBiasMetrics": GcBiasMetrics,
            "HsMetrics": HsMetrics,
            "InsertSizeMetrics": InsertSizeMetrics,
            "MarkDuplicates": MarkDuplicates,
            "OxoGMetrics": OxoGMetrics,
            "QualityByCycleMetrics": QualityByCycleMetrics,
            "QualityScoreDistributionMetrics": QualityScoreDistributionMetrics,
            "QualityYieldMetrics": QualityYieldMetrics,
            "RnaSeqMetrics": RnaSeqMetrics,
            "RrbsSummaryMetrics": RrbsSummaryMetrics,
            "TargetedPcrMetrics": TargetedPcrMetrics,
            "VariantCallingMetrics": VariantCallingMetrics,
            "ValidateSamFile": ValidateSamFile,
            "WgsMetrics": WgsMetrics,
            "IlluminaBasecallingMetrics": IlluminaBasecallingMetrics,
            "IlluminaLaneMetrics": IlluminaLaneMetrics,
            "ExtractIlluminaBarcodes": ExtractIlluminaBarcodes,
            "MarkIlluminaAdapters": MarkIlluminaAdapters,
        }.items():
            func = getattr(mod, "parse_reports", None)
            if func is not None:
                n[tool] = func(self)
                if n[tool] > 0:
                    log.info(f"Found {n[tool]} {tool} reports")

        # Exit if we didn't find anything
        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound



""" Core MultiQC module to parse output from custom script output """


import base64
import json
import logging
import os
import re
from collections import defaultdict

import yaml

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, beeswarm, heatmap, linegraph, scatter, table
from multiqc.utils import report

# Initialise the logger
log = logging.getLogger(__name__)


def custom_module_classes():
    """
    MultiQC Custom Content class. This module does a lot of different
    things depending on the input and is as flexible as possible.

    NB: THIS IS TOTALLY DIFFERENT TO ALL OTHER MODULES
    """

    # Dict to hold parsed data. Each key should contain a custom data type
    # e.g. output from a particular script. Note that this script may pick
    # up many different types of data from many different sources.
    # Second level keys should be 'config' and 'data'. Data key should then
    # contain sample names, and finally data.
    cust_mods = defaultdict(lambda: defaultdict(lambda: dict()))

    # Dictionary to hold search patterns - start with those defined in the config
    search_patterns = ["custom_content"]

    # First - find files using patterns described in the config
    config_data = getattr(config, "custom_data", {})
    mod_cust_config = {}
    for k, f in config_data.items():
        # Check that we have a dictionary
        if not isinstance(f, dict):
            log.debug(f"config.custom_data row was not a dictionary: {k}")
            continue
        c_id = f.get("id", k)

        # Data supplied in with config (e.g. from a multiqc_config.yaml file in working directory)
        if "data" in f:
            try:
                cust_mods[c_id]["data"].update(f["data"])
            except ValueError:
                # HTML plot type doesn't have a data sample-id key, so just take the whole chunk of data
                cust_mods[c_id]["data"] = f["data"]
            cust_mods[c_id]["config"].update({k: v for k, v in f.items() if k != "data"})
            cust_mods[c_id]["config"]["id"] = cust_mods[c_id]["config"].get("id", c_id)
            continue

        # Custom Content ID has search patterns in the config
        if c_id in report.files:
            cust_mods[c_id]["config"] = f
            cust_mods[c_id]["config"]["id"] = cust_mods[c_id]["config"].get("id", c_id)
            search_patterns.append(c_id)
            continue

        # Must just be configuration for a separate custom-content class
        mod_cust_config[c_id] = f

    # Now go through each of the file search patterns
    bm = BaseMultiqcModule(name="Custom content", anchor="custom_content")
    for k in search_patterns:
        num_sp_found_files = 0
        for f in bm.find_log_files(k):
            num_sp_found_files += 1
            # Handle any exception without messing up for remaining custom content files
            try:
                f_extension = os.path.splitext(f["fn"])[1]

                # YAML and JSON files are the easiest
                parsed_data = None
                if f_extension == ".yaml" or f_extension == ".yml":
                    try:
                        parsed_data = yaml.safe_load(f["f"])
                    except Exception as e:
                        log.warning(f"Error parsing YAML file '{f['fn']}' (probably invalid YAML)")
                        log.debug(f"YAML error: {e}", exc_info=True)
                        break
                    parsed_data["id"] = parsed_data.get("id", f["s_name"])
                elif f_extension == ".json":
                    try:
                        parsed_data = json.loads(f["f"])
                    except Exception as e:
                        log.warning(f"Error parsing JSON file '{f['fn']}' (probably invalid JSON)")
                        log.warning(f"JSON error: {e}")
                        break
                    parsed_data["id"] = parsed_data.get("id", f["s_name"])
                elif f_extension == ".png" or f_extension == ".jpeg" or f_extension == ".jpg":
                    image_string = base64.b64encode(f["f"].read()).decode("utf-8")
                    image_format = "png" if f_extension == ".png" else "jpg"
                    img_html = (
                        '<div class="mqc-custom-content-image"><img src="data:image/{};base64,{}" /></div>'.format(
                            image_format, image_string
                        )
                    )
                    parsed_data = {
                        "id": f["s_name"],
                        "plot_type": "image",
                        "section_name": f["s_name"].replace("_", " ").replace("-", " ").replace(".", " "),
                        "data": img_html,
                    }
                    # If the search pattern 'k' has an associated custom content section config, use it
                    parsed_data.update(cust_mods.get(k, {}).get("config", {}))
                elif f_extension == ".html":
                    parsed_data = {"id": f["s_name"], "plot_type": "html", "data": f["f"]}
                    parsed_data.update(_find_html_file_header(f))

                if parsed_data is not None:
                    if isinstance(parsed_data.get("data"), dict):
                        # Run sample-name cleaning on the data keys
                        parsed_data["data"] = {bm.clean_s_name(k, f): v for k, v in parsed_data["data"].items()}

                    c_id = parsed_data.get("id", k)
                    if len(parsed_data.get("data", {})) > 0:
                        if isinstance(parsed_data["data"], dict):
                            cust_mods[c_id]["data"].update(parsed_data["data"])
                        else:
                            cust_mods[c_id]["data"] = parsed_data["data"]
                        cust_mods[c_id]["config"].update({j: k for j, k in parsed_data.items() if j != "data"})
                    else:
                        log.warning(f"No data found in {f['fn']}")

                # txt, csv, tsv etc
                else:
                    # Look for configuration details in the header
                    m_config = _find_file_header(f)
                    s_name = None
                    if m_config is not None:
                        c_id = m_config.get("id", k)
                        # Update the base config with anything parsed from the file
                        b_config = cust_mods.get(c_id, {}).get("config", {})
                        b_config.update(m_config)
                        # Now set the module config to the merged dict
                        m_config = dict(b_config)
                        s_name = m_config.get("sample_name")
                    else:
                        c_id = k
                        m_config = cust_mods.get(c_id, {}).get("config", {})

                    # Guess sample name if not given
                    if s_name is None:
                        s_name = f["s_name"]

                    # Guess c_id if no information known
                    if k == "custom_content":
                        c_id = s_name
                        if not m_config.get("id"):
                            m_config["id"] = c_id

                    # Merge with config from a MultiQC config file if we have it
                    m_config.update(mod_cust_config.get(c_id, {}))

                    # Add information about the file to the config dict
                    if "files" not in m_config:
                        m_config["files"] = dict()
                    m_config["files"].update({s_name: {"fn": f["fn"], "root": f["root"]}})

                    # Guess file format if not given
                    if m_config.get("file_format") is None:
                        m_config["file_format"] = _guess_file_format(f)
                    # Parse data
                    try:
                        parsed_data, conf = _parse_txt(f, m_config)
                        if parsed_data is None or len(parsed_data) == 0:
                            log.warning(f"Not able to parse custom data in {f['fn']}")
                        else:
                            # Did we get a new section id from the file?
                            if conf.get("id") is not None:
                                c_id = conf.get("id")
                            # heatmap - special data type
                            if isinstance(parsed_data, list):
                                cust_mods[c_id]["data"] = parsed_data
                            elif conf.get("plot_type") == "html":
                                cust_mods[c_id]["data"] = parsed_data
                            else:
                                cust_mods[c_id]["data"].update(parsed_data)
                            cust_mods[c_id]["config"].update(conf)
                    except (IndexError, AttributeError, TypeError):
                        log.error(f"Unexpected parsing error for {f['fn']}", exc_info=True)
                        raise  # testing
            except Exception as e:
                log.error(f"Uncaught exception raised for file '{f['fn']}'")
                log.exception(e)

        # Give log message if no files found for search pattern
        if num_sp_found_files == 0 and k != "custom_content":
            log.debug(f"No samples found: custom content ({k})")

    # Filter to strip out ignored sample names
    for k in cust_mods:
        cust_mods[k]["data"] = bm.ignore_samples(cust_mods[k]["data"])

    # Remove any configs that have no data
    remove_cids = [k for k in cust_mods if len(cust_mods[k]["data"]) == 0]
    for k in remove_cids:
        del cust_mods[k]

    if len(cust_mods) == 0:
        raise ModuleNoSamplesFound

    # Go through each data type
    parsed_modules = dict()
    for c_id, mod in cust_mods.items():
        # General Stats
        if mod["config"].get("plot_type") == "generalstats":
            gsheaders = mod["config"].get("pconfig")
            if gsheaders is None:
                headers = set()
                for d in mod["data"].values():
                    headers.update(d.keys())
                headers = list(headers)
                headers.sort()
                gsheaders = dict()
                for h in headers:
                    gsheaders[h] = dict()

            # Headers is a list of dicts
            if isinstance(gsheaders, list):
                gsheaders_dict = dict()
                for gsheader in gsheaders:
                    for col_id, col_data in gsheader.items():
                        gsheaders_dict[col_id] = col_data
                gsheaders = gsheaders_dict

            # Add namespace and description if not specified
            for m_id in gsheaders:
                if "namespace" not in gsheaders[m_id]:
                    gsheaders[m_id]["namespace"] = mod["config"].get("namespace", c_id)
            log.info(f"{c_id}: Found {len(mod['data'])} General Statistics columns")
            bm.general_stats_addcols(mod["data"], gsheaders)

        # Initialise this new module class and append to list
        else:
            # Is this file asking to be a sub-section under a parent section?
            mod_id = mod["config"].get("parent_id", c_id)
            # If we have any custom configuration from a MultiQC config file, update here
            # This is done earlier for tsv files too, but we do it here so that it overwrites what was in the file
            if mod_id in mod_cust_config:
                mod["config"].update(mod_cust_config[mod_id])
            # We've not seen this module section before (normal for most custom content)
            if mod_id not in parsed_modules:
                parsed_modules[mod_id] = MultiqcModule(mod_id, mod)
            else:
                # New sub-section
                parsed_modules[mod_id].update_init(c_id, mod)
            parsed_modules[mod_id].add_cc_section(c_id, mod)
            if mod["config"].get("plot_type") == "html":
                log.info(f"{c_id}: Found 1 sample (html)")
            elif mod["config"].get("plot_type") == "image":
                log.info(f"{c_id}: Found 1 sample (image)")
            else:
                log.info(f"{c_id}: Found {len(mod['data'])} samples ({mod['config'].get('plot_type')})")

    # Sort sections if we have a config option for order
    mod_order = getattr(config, "custom_content", {}).get("order", [])
    sorted_modules = [parsed_mod for parsed_mod in parsed_modules.values() if parsed_mod.anchor not in mod_order]
    sorted_modules.extend(
        [parsed_mod for mod_id in mod_order for parsed_mod in parsed_modules.values() if parsed_mod.anchor == mod_id]
    )

    # If we only have General Stats columns then there are no module outputs
    if len(sorted_modules) == 0:
        if mod["config"].get("plot_type") == "generalstats":
            sorted_modules = [bm]
        else:
            raise ModuleNoSamplesFound

    return sorted_modules


class MultiqcModule(BaseMultiqcModule):
    """Module class, used for each custom content type"""

    def __init__(self, c_id, mod):
        modname = c_id.replace("_", " ").title()
        mod_info = mod["config"].get("description")
        if "parent_name" in mod["config"]:
            modname = mod["config"]["parent_name"]
            mod_info = mod["config"].get("parent_description")
        elif "section_name" in mod["config"]:
            modname = mod["config"]["section_name"]
        if modname == "" or modname is None:
            modname = "Custom Content"

        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name=modname,
            anchor=mod["config"].get("section_anchor", c_id),
            href=mod["config"].get("section_href"),
            info=mod_info,
            extra=mod["config"].get("extra"),
            # No DOI here.. // doi=
        )

        # Don't repeat the Custom Content name in the subtext
        if self.info or self.extra:
            self.intro = f"<p>{self.info}</p>{self.extra}"

    def update_init(self, c_id, mod):
        """
        This function runs when we have already initialised a module
        and this is a subsequent file that will be another subsection.

        So most info should already be initialised properly.
        We check if anything is empty and if we have it set in this file
        we set it here.
        """

        if self.info is None or self.info == "":
            self.info = mod["config"].get("parent_description")
        if self.extra is None or self.info == "":
            self.extra = mod["config"].get("extra", None)
        # This needs overwriting again as it has already run on init
        if self.info or self.extra:
            self.intro = f"<p>{self.info}</p>{self.extra}"

    def add_cc_section(self, c_id, mod):
        section_name = mod["config"].get("section_name", c_id.replace("_", " ").title())
        if section_name == "" or section_name is None:
            section_name = "Custom Content"

        section_description = mod["config"].get("description", "")

        pconfig = mod["config"].get("pconfig", {})
        if pconfig.get("id") is None:
            pconfig["id"] = f"{c_id}-plot"
        if pconfig.get("title") is None:
            pconfig["title"] = section_name

        plot = None
        content = None

        if not isinstance(mod["data"], list):
            mod["data"] = [mod["data"]]

        for i, ds in enumerate(mod["data"]):
            # Save the data if it's not a html string
            if not isinstance(ds, str):
                did = pconfig["id"]
                if i > 0:
                    did = f"{did}_{i}"
                self.write_data_file(ds, f"multiqc_{did}")
                pconfig["save_data_file"] = False

        # Try to coerce x-axis to numeric
        if mod["config"].get("plot_type") in ["linegraph", "scatter"]:
            try:
                mod["data"] = [{k: {float(x): v[x] for x in v} for k, v in ds.items()} for ds in mod["data"]]
            except ValueError:
                pass

        # Table
        if mod["config"].get("plot_type") == "table":
            pconfig["sortRows"] = pconfig.get("sortRows", False)
            headers = mod["config"].get("headers")
            plot = table.plot(mod["data"], headers, pconfig)

        # Bar plot
        elif mod["config"].get("plot_type") == "bargraph":
            mod["data"] = [{str(k): v for k, v in ds.items()} for ds in mod["data"]]
            plot = bargraph.plot(mod["data"], mod["config"].get("categories"), pconfig)

        # Line plot
        elif mod["config"].get("plot_type") == "linegraph":
            plot = linegraph.plot(mod["data"], pconfig)

        # Scatter plot
        elif mod["config"].get("plot_type") == "scatter":
            plot = scatter.plot(mod["data"], pconfig)

        # Heatmap
        elif mod["config"].get("plot_type") == "heatmap":
            plot = heatmap.plot(mod["data"], mod["config"].get("xcats"), mod["config"].get("ycats"), pconfig)

        # Beeswarm plot
        elif mod["config"].get("plot_type") == "beeswarm":
            plot = beeswarm.plot(mod["data"], pconfig)

        # Raw HTML
        elif mod["config"].get("plot_type") == "html":
            if len(mod["data"]) > 1:
                log.warning(f"HTML plot type found with more than one dataset in {c_id}")
            content = mod["data"][0]

        # Raw image file as html
        elif mod["config"].get("plot_type") == "image":
            if len(mod["data"]) > 1:
                log.warning(f"Image plot type found with more than one dataset in {c_id}")
            content = mod["data"][0]

        # Not supplied
        elif mod["config"].get("plot_type") is None:
            log.warning(f"Plot type not found for content ID '{c_id}'")

        # Not recognised
        else:
            log.warning(
                "Error - custom content plot type '{}' not recognised for content ID {}".format(
                    mod["config"].get("plot_type"), c_id
                )
            )

        # Don't use exactly the same title / description text as the main module
        if section_name == self.name:
            section_name = None
        if self.info and section_description.strip(".") == self.info.strip("."):
            section_description = ""

        self.add_section(name=section_name, anchor=c_id, description=section_description, plot=plot, content=content)


def _find_file_header(f):
    # Collect commented out header lines
    hlines = []
    for line in f["f"].splitlines():
        if line.startswith("#"):
            hlines.append(line[1:])
    if len(hlines) == 0:
        return None
    hconfig = None
    try:
        hconfig = yaml.safe_load("\n".join(hlines))
        assert isinstance(hconfig, dict)
    except yaml.YAMLError as e:
        log.warning(f"Could not parse comment file header for MultiQC custom content: {f['fn']}")
        log.debug(e)
    except AssertionError:
        log.debug(f"Custom Content comment file header looked wrong: {hconfig}")
    else:
        return hconfig


def _find_html_file_header(f):
    """Look for a HTML comment config at the start of a custom content HTML file"""
    if f["f"].lstrip().startswith("<!--"):
        match = re.search(r"^\<\!\-\-((?:.|\n|\r)*?)-->", f["f"].lstrip())
        if match:
            comment = match.group(1)
            if comment:
                try:
                    return yaml.load(comment, Loader=yaml.SafeLoader)
                except Exception as e:
                    log.debug(f"Found Custom Content HTML comment, but couldn't load as YAML: {e}", exc_info=True)
                    log.debug(f"Comment:\n{comment}")
    return {}


def _guess_file_format(f):
    """
    Tries to guess file format, first based on file extension (csv / tsv),
    then by looking for common column separators in the first 10 non-commented lines.
    Splits by tab / comma / space and counts resulting number of columns. Finds the most
    common column count, then compared how many lines had this number.
    e.g. if tab, all 10 lines should have x columns when split by tab.
    Returns: csv | tsv | spaces   (spaces by default if all else fails)
    """
    filename, file_extension = os.path.splitext(f["fn"])
    tabs = []
    commas = []
    spaces = []
    j = 0
    for line in f["f"].splitlines():
        if not line.startswith("#"):
            j += 1
            tabs.append(len(line.split("\t")))
            commas.append(len(line.split(",")))
            spaces.append(len(line.split()))
        if j == 10:
            break
    tab_mode = max(set(tabs), key=tabs.count)
    commas_mode = max(set(commas), key=commas.count)
    spaces_mode = max(set(spaces), key=spaces.count)
    tab_lc = tabs.count(tab_mode) if tab_mode > 1 else 0
    commas_lc = commas.count(commas_mode) if commas_mode > 1 else 0
    spaces_lc = spaces.count(spaces_mode) if spaces_mode > 1 else 0
    if tab_lc == j:
        return "tsv"
    elif commas_lc == j:
        return "csv"
    else:
        if tab_lc > commas_lc and tab_lc > spaces_lc:
            return "tsv"
        elif commas_lc > tab_lc and commas_lc > spaces_lc:
            return "csv"
        elif spaces_lc > tab_lc and spaces_lc > commas_lc:
            return "spaces"
        else:
            if tab_mode == commas_lc and tab_mode > spaces_lc:
                if tab_mode > commas_mode:
                    return "tsv"
                else:
                    return "csv"
    return "spaces"


def _parse_txt(f, conf):
    # Split the data into a list of lists by column
    sep = None
    if conf["file_format"] == "csv":
        sep = ","
    if conf["file_format"] == "tsv":
        sep = "\t"
    lines = f["f"].splitlines()
    d = []

    # Check for special case - HTML
    if conf.get("plot_type") == "html":
        for line in lines:
            if line and not line.startswith("#"):
                d.append(line)
        return "\n".join(d), conf

    # Not HTML, need to parse data
    ncols = None
    for line in lines:
        if line and not line.startswith("#"):
            sections = line.split(sep)
            d.append(sections)
            if ncols is None:
                ncols = len(sections)
            elif ncols != len(sections):
                log.warning(f"Inconsistent number of columns found in {f['fn']}! Skipping..")
                return None, conf

    # Convert values to floats if we can
    first_row_str = 0
    for i, line in enumerate(d):
        for j, v in enumerate(line):
            try:
                v = float(v)
            except ValueError:
                pass
            if isinstance(v, str):
                if (v.startswith('"') and v.endswith('"')) or (v.startswith("'") and v.endswith("'")):
                    v = v[1:-1]
                # Count strings in first row (header?)
                if i == 0:
                    first_row_str += 1
            d[i][j] = v

    all_numeric = all([isinstance(line, float) for i in range(1, len(d)) for line in d[i][1:]])

    # General stat info files - expected to have at least 2 rows (first row always being the header)
    # and have at least 2 columns (first column always being sample name)
    if conf.get("plot_type") == "generalstats" and len(d) >= 2 and ncols >= 2:
        data = defaultdict(dict)
        for i, line in enumerate(d[1:], 1):
            for j, v in enumerate(line[1:], 1):
                data[line[0]][d[0][j]] = v
        return data, conf

    # Heatmap: Number of headers == number of lines
    if conf.get("plot_type") is None and first_row_str == len(lines) and all_numeric:
        conf["plot_type"] = "heatmap"
    if conf.get("plot_type") == "heatmap":
        conf["xcats"] = d[0][1:]
        conf["ycats"] = [s[0] for s in d[1:]]
        data = [s[1:] for s in d[1:]]
        return data, conf

    # Header row of strings, or configured as table
    if first_row_str == len(d[0]) or conf.get("plot_type") == "table":
        data = dict()
        for s in d[1:]:
            data[s[0]] = dict()
            for i, v in enumerate(s[1:]):
                cat = str(d[0][i + 1])
                data[s[0]][cat] = v
        # Bar graph or table - if numeric data, go for bar graph
        if conf.get("plot_type") is None:
            allfloats = True
            for r in d[1:]:
                for v in r[1:]:
                    allfloats = allfloats and isinstance(v, float)
            if allfloats:
                conf["plot_type"] = "bargraph"
            else:
                conf["plot_type"] = "table"
        # Set table col_1 header
        if conf.get("plot_type") == "table" and d[0][0].strip() != "":
            conf["pconfig"] = conf.get("pconfig", {})
            if not conf["pconfig"].get("col1_header"):
                conf["pconfig"]["col1_header"] = d[0][0].strip()
        # Return parsed data
        if conf.get("plot_type") == "bargraph" or conf.get("plot_type") == "table":
            return data, conf
        else:
            data = dict()  # reset

    # Scatter plot: First row is  str : num : num
    if (
        conf.get("plot_type") is None
        and len(d[0]) == 3
        and not isinstance(d[0][0], float)
        and isinstance(d[0][1], float)
        and isinstance(d[0][2], float)
    ):
        conf["plot_type"] = "scatter"

    if conf.get("plot_type") == "scatter":
        data = dict()
        for s in d:
            try:
                data[s[0]] = {"x": float(s[1]), "y": float(s[2])}
            except (IndexError, ValueError):
                pass
        return data, conf

    # Single sample line / bar graph - first row has two columns
    if len(d[0]) == 2:
        # Line graph - num : num
        if conf.get("plot_type") is None and isinstance(d[0][0], float) and isinstance(d[0][1], float):
            conf["plot_type"] = "linegraph"
        # Bar graph - str : num
        if conf.get("plot_type") is None and not isinstance(d[0][0], float) and isinstance(d[0][1], float):
            conf["plot_type"] = "bargraph"

        # Data structure is the same
        if conf.get("plot_type") == "linegraph" or conf.get("plot_type") == "bargraph":
            # Set section id based on directory if not known
            if conf.get("id") is None:
                conf["id"] = os.path.basename(f["root"])
            data = dict()
            for s in d:
                data[s[0]] = s[1]
            return {f["s_name"]: data}, conf

    # Multi-sample line graph: No header row, str : lots of num columns
    if conf.get("plot_type") is None and len(d[0]) > 4 and all_numeric:
        conf["plot_type"] = "linegraph"

    if conf.get("plot_type") == "linegraph":
        data = dict()
        # If the first row has no header, use it as axis labels
        x_labels = []
        if d[0][0].strip() == "":
            x_labels = d.pop(0)[1:]
        # Use 1..n range for x values
        for s in d:
            data[s[0]] = dict()
            for i, v in enumerate(s[1:]):
                try:
                    x_val = x_labels[i]
                    try:
                        x_val = float(x_val)
                    except ValueError:
                        pass
                except IndexError:
                    x_val = i + 1
                data[s[0]][x_val] = v
        return data, conf

    # Got to the end and haven't returned. It's a mystery, capn'!
    log.debug(
        f"Not able to figure out a plot type for '{f['fn']}' "
        + "plot type = {}, all numeric = {}, first row str = {}".format(
            conf.get("plot_type"), all_numeric, first_row_str
        )
    )
    return None, conf
